# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2016 Red Hat, Inc.
# Copyright 2021-2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import random
import time
import json
import re
import urllib.parse
import socketserver
import datetime
import threading
import string
import copy
import http.server

from zuul.lib.logutil import get_annotated_logger
from zuul.driver.git import gitwatcher
from zuul.driver.gerrit import (
    gerritconnection,
    gerritreporter,
)
from tests.util import FIXTURE_DIR, random_sha1

import git


class GerritChangeReference(git.Reference):
    _common_path_default = "refs/changes"
    _points_to_commits_only = True


class FakeGerritChange(object):
    categories = {'Approved': ('Approved', -1, 1),
                  'Code-Review': ('Code-Review', -2, 2),
                  'Verified': ('Verified', -2, 2)}

    def __init__(self, gerrit, number, project, branch, subject,
                 status='NEW', upstream_root=None, files={},
                 parent=None, merge_parents=None, merge_files=None,
                 merge_parent_files=None,
                 topic=None, empty=False):
        self.cherry_pick = False
        self.source_hostname = gerrit.canonical_hostname
        self.gerrit_baseurl = gerrit.baseurl
        self.reported = 0
        self.notify = None
        self.queried = 0
        self.patchsets = []
        self.number = number
        self.project = project
        self.branch = branch
        self.subject = subject
        self.latest_patchset = 0
        self.depends_on_change = None
        self.depends_on_patchset = None
        self.needed_by_changes = []
        self.fail_merge = False
        self.messages = []
        self.comments = []
        self.checks = {}
        self.checks_history = []
        self.submit_requirements = []
        self.data = {
            'branch': branch,
            'comments': self.comments,
            'commitMessage': subject,
            'createdOn': time.time(),
            'id': 'I' + random_sha1(),
            'lastUpdated': time.time(),
            'number': str(number),
            'open': status == 'NEW',
            'owner': {'email': 'user@example.com',
                      'name': 'User Name',
                      'username': 'username'},
            'patchSets': self.patchsets,
            'project': project,
            'status': status,
            'subject': subject,
            'submitRecords': [],
            'hashtags': [],
            'url': '%s/%s' % (self.gerrit_baseurl.rstrip('/'), number)}

        if topic:
            self.data['topic'] = topic
        self.upstream_root = upstream_root
        self.merge_parent_files = merge_parent_files
        if merge_parents:
            self.addMergePatchset(parents=merge_parents,
                                  merge_files=merge_files)
        else:
            self.addPatchset(files=files, parent=parent, empty=empty)
        if merge_parents:
            self.data['parents'] = merge_parents
        elif parent:
            self.data['parents'] = [parent]
        self.data['submitRecords'] = self.getSubmitRecords()
        self.open = status == 'NEW'

    def __setstate__(self, state):
        self.__dict__.update(state)
        # Only needed for upgrade test when adding new attribute.
        if not hasattr(self, 'cherry_pick'):
            self.cherry_pick = False

    def addFakeChangeToRepo(self, msg, files, large, parent):
        path = os.path.join(self.upstream_root, self.project)
        repo = git.Repo(path)
        if parent is None:
            parent = 'refs/tags/init'
        ref = GerritChangeReference.create(
            repo, '%s/%s/%s' % (str(self.number).zfill(2)[-2:],
                                self.number,
                                self.latest_patchset),
            parent)
        repo.head.reference = ref
        repo.head.reset(working_tree=True)
        repo.git.clean('-x', '-f', '-d')

        path = os.path.join(self.upstream_root, self.project)
        if not large:
            for fn, content in files.items():
                fn = os.path.join(path, fn)
                if content is None:
                    os.unlink(fn)
                    repo.index.remove([fn])
                else:
                    d = os.path.dirname(fn)
                    if not os.path.exists(d):
                        os.makedirs(d)
                    with open(fn, 'w') as f:
                        f.write(content)
                    repo.index.add([fn])
        else:
            for fni in range(100):
                fn = os.path.join(path, str(fni))
                f = open(fn, 'w')
                for ci in range(4096):
                    f.write(random.choice(string.printable))
                f.close()
                repo.index.add([fn])

        r = repo.index.commit(msg)
        repo.head.reference = 'master'
        repo.head.reset(working_tree=True)
        repo.git.clean('-x', '-f', '-d')
        repo.heads['master'].checkout()
        return r

    def addFakeMergeCommitChangeToRepo(self, msg, parents):
        path = os.path.join(self.upstream_root, self.project)
        repo = git.Repo(path)
        ref = GerritChangeReference.create(
            repo, '%s/%s/%s' % (str(self.number).zfill(2)[-2:],
                                self.number,
                                self.latest_patchset),
            parents[0])
        repo.head.reference = ref
        repo.head.reset(working_tree=True)
        repo.git.clean('-x', '-f', '-d')

        repo.index.merge_tree(parents[1])
        parent_commits = [repo.commit(p) for p in parents]
        r = repo.index.commit(msg, parent_commits=parent_commits)

        repo.head.reference = 'master'
        repo.head.reset(working_tree=True)
        repo.git.clean('-x', '-f', '-d')
        repo.heads['master'].checkout()
        return r

    def addPatchset(self, files=None, large=False, parent=None, empty=False):
        self.latest_patchset += 1
        if empty:
            files = {}
        elif not files:
            fn = '%s-%s' % (self.branch.replace('/', '_'), self.number)
            data = ("test %s %s %s\n" %
                    (self.branch, self.number, self.latest_patchset))
            files = {fn: data}
        msg = self.subject + '-' + str(self.latest_patchset)
        c = self.addFakeChangeToRepo(msg, files, large, parent)
        ps_files = [{'file': '/COMMIT_MSG',
                     'type': 'ADDED'},
                    {'file': 'README',
                     'type': 'MODIFIED'}]
        for f in files:
            ps_files.append({'file': f, 'type': 'ADDED'})
        d = {'approvals': [],
             'createdOn': time.time(),
             'files': ps_files,
             'number': str(self.latest_patchset),
             'ref': 'refs/changes/%s/%s/%s' % (str(self.number).zfill(2)[-2:],
                                               self.number,
                                               self.latest_patchset),
             'revision': c.hexsha,
             'uploader': {'email': 'user@example.com',
                          'name': 'User name',
                          'username': 'user'}}
        self.data['currentPatchSet'] = d
        self.patchsets.append(d)
        self.data['submitRecords'] = self.getSubmitRecords()

    def addCherryPickPatchset(self):
        path = os.path.join(self.upstream_root, self.project)
        repo = git.Repo(path)
        self.latest_patchset += 1
        prev = self.patchsets[-1]
        d = {
            'approvals': prev['approvals'],
            'createdOn': time.time(),
            'files': prev['files'],
            'number': str(self.latest_patchset),
            'ref': 'refs/changes/%s/%s/%s' % (str(self.number).zfill(2)[-2:],
                                              self.number,
                                              self.latest_patchset),
            'revision': repo.heads[self.branch].commit.hexsha,
            'uploader': prev['uploader'],
        }
        self.data['currentPatchSet'] = d
        self.patchsets.append(d)
        self.data['submitRecords'] = self.getSubmitRecords()

    def addMergePatchset(self, parents, merge_files=None):
        self.latest_patchset += 1
        if not merge_files:
            merge_files = []
        msg = self.subject + '-' + str(self.latest_patchset)
        c = self.addFakeMergeCommitChangeToRepo(msg, parents)
        ps_files = [{'file': '/COMMIT_MSG',
                     'type': 'ADDED'},
                    {'file': '/MERGE_LIST',
                     'type': 'ADDED'}]
        for f in merge_files:
            ps_files.append({'file': f, 'type': 'ADDED'})
        d = {'approvals': [],
             'createdOn': time.time(),
             'files': ps_files,
             'number': str(self.latest_patchset),
             'ref': 'refs/changes/%s/%s/%s' % (str(self.number).zfill(2)[-2:],
                                               self.number,
                                               self.latest_patchset),
             'revision': c.hexsha,
             'uploader': {'email': 'user@example.com',
                          'name': 'User name',
                          'username': 'user'}}
        self.data['currentPatchSet'] = d
        self.patchsets.append(d)
        self.data['submitRecords'] = self.getSubmitRecords()

    def setCheck(self, checker, reset=False, **kw):
        if reset:
            self.checks[checker] = {'state': 'NOT_STARTED',
                                    'created': str(datetime.datetime.now())}
        chk = self.checks.setdefault(checker, {})
        chk['updated'] = str(datetime.datetime.now())
        for (key, default) in [
                ('state', None),
                ('repository', self.project),
                ('change_number', self.number),
                ('patch_set_id', self.latest_patchset),
                ('checker_uuid', checker),
                ('message', None),
                ('url', None),
                ('started', None),
                ('finished', None),
        ]:
            val = kw.get(key, chk.get(key, default))
            if val is not None:
                chk[key] = val
            elif key in chk:
                del chk[key]
        self.checks_history.append(copy.deepcopy(self.checks))

    def addComment(self, filename, line, message, name, email, username,
                   comment_range=None):
        comment = {
            'file': filename,
            'line': int(line),
            'reviewer': {
                'name': name,
                'email': email,
                'username': username,
            },
            'message': message,
        }
        if comment_range:
            comment['range'] = comment_range
        self.comments.append(comment)

    def getPatchsetCreatedEvent(self, patchset):
        event = {"type": "patchset-created",
                 "change": {"project": self.project,
                            "branch": self.branch,
                            "id": "I5459869c07352a31bfb1e7a8cac379cabfcb25af",
                            "number": str(self.number),
                            "subject": self.subject,
                            "owner": {"name": "User Name"},
                            "url": "https://hostname/3"},
                 "patchSet": self.patchsets[patchset - 1],
                 "uploader": {"name": "User Name"}}
        return event

    def getChangeRestoredEvent(self):
        event = {"type": "change-restored",
                 "change": {"project": self.project,
                            "branch": self.branch,
                            "id": "I5459869c07352a31bfb1e7a8cac379cabfcb25af",
                            "number": str(self.number),
                            "subject": self.subject,
                            "owner": {"name": "User Name"},
                            "url": "https://hostname/3"},
                 "restorer": {"name": "User Name"},
                 "patchSet": self.patchsets[-1],
                 "reason": ""}
        return event

    def getChangeAbandonedEvent(self):
        event = {"type": "change-abandoned",
                 "change": {"project": self.project,
                            "branch": self.branch,
                            "id": "I5459869c07352a31bfb1e7a8cac379cabfcb25af",
                            "number": str(self.number),
                            "subject": self.subject,
                            "owner": {"name": "User Name"},
                            "url": "https://hostname/3"},
                 "abandoner": {"name": "User Name"},
                 "patchSet": self.patchsets[-1],
                 "reason": ""}
        return event

    def getChangeCommentEvent(self, patchset, comment=None,
                              patchsetcomment=None):
        if comment is None and patchsetcomment is None:
            comment = "Patch Set %d:\n\nThis is a comment" % patchset
        elif comment:
            comment = "Patch Set %d:\n\n%s" % (patchset, comment)
        else:  # patchsetcomment is not None
            comment = "Patch Set %d:\n\n(1 comment)" % patchset

        commentevent = {"comment": comment}
        if patchsetcomment:
            commentevent.update(
                {'patchSetComments':
                    {"/PATCHSET_LEVEL": [{"message": patchsetcomment}]}
                }
            )

        event = {"type": "comment-added",
                 "change": {"project": self.project,
                            "branch": self.branch,
                            "id": "I5459869c07352a31bfb1e7a8cac379cabfcb25af",
                            "number": str(self.number),
                            "subject": self.subject,
                            "owner": {"name": "User Name"},
                            "url": "https://hostname/3"},
                 "patchSet": self.patchsets[patchset - 1],
                 "author": {"name": "User Name"},
                 "approvals": [{"type": "Code-Review",
                                "description": "Code-Review",
                                "value": "0"}]}
        event.update(commentevent)
        return event

    def getChangeMergedEvent(self):
        event = {"submitter": {"name": "Jenkins",
                               "username": "jenkins"},
                 "newRev": self.patchsets[-1]['revision'],
                 "patchSet": self.patchsets[-1],
                 "change": self.data,
                 "type": "change-merged",
                 "eventCreatedOn": 1487613810}
        return event

    def getRefUpdatedEvent(self):
        path = os.path.join(self.upstream_root, self.project)
        repo = git.Repo(path)
        oldrev = repo.heads[self.branch].commit.hexsha

        event = {
            "type": "ref-updated",
            "submitter": {
                "name": "User Name",
            },
            "refUpdate": {
                "oldRev": oldrev,
                "newRev": self.patchsets[-1]['revision'],
                "refName": self.branch,
                "project": self.project,
            }
        }
        return event

    def getHashtagsChangedEvent(self, added=None, removed=None):
        event = {
            'type': 'hashtags-changed',
            'change': {'branch': self.branch,
                       'commitMessage': self.data['commitMessage'],
                       'createdOn': 1689442009,
                       'id': 'I254acfc54f9942394ff924a806cd87c70cec2f4d',
                       'number': int(self.number),
                       'owner': self.data['owner'],
                       'project': self.project,
                       'status': self.data['status'],
                       'subject': self.subject,
                       'url': 'https://hostname/3'},
            'changeKey': {'id': 'I254acfc54f9942394ff924a806cd87c70cec2f4d'},
            'editor': {'email': 'user@example.com',
                       'name': 'User Name',
                       'username': 'user'},
            'eventCreatedOn': 1701711038,
            'project': self.project,
            'refName': self.branch,
        }
        if added:
            event['added'] = added
        if removed:
            event['removed'] = removed
        return event

    def addApproval(self, category, value, username='reviewer_john',
                    granted_on=None, message='', tag=None, old_value=None):
        if not granted_on:
            granted_on = time.time()
        approval = {
            'description': self.categories[category][0],
            'type': category,
            'value': str(value),
            'by': {
                'username': username,
                'email': username + '@example.com',
            },
            'grantedOn': int(granted_on),
            '__tag': tag,  # Not available in ssh api
        }
        if old_value is not None:
            approval['oldValue'] = str(old_value)
        for i, x in enumerate(self.patchsets[-1]['approvals'][:]):
            if x['by']['username'] == username and x['type'] == category:
                del self.patchsets[-1]['approvals'][i]
        self.patchsets[-1]['approvals'].append(approval)
        event = {'approvals': [approval],
                 'author': {'email': 'author@example.com',
                            'name': 'Patchset Author',
                            'username': 'author_phil'},
                 'change': {'branch': self.branch,
                            'id': 'Iaa69c46accf97d0598111724a38250ae76a22c87',
                            'number': str(self.number),
                            'owner': {'email': 'owner@example.com',
                                      'name': 'Change Owner',
                                      'username': 'owner_jane'},
                            'project': self.project,
                            'subject': self.subject,
                            'url': 'https://hostname/459'},
                 'comment': message,
                 'patchSet': self.patchsets[-1],
                 'type': 'comment-added'}
        if 'topic' in self.data:
            event['change']['topic'] = self.data['topic']

        self.data['submitRecords'] = self.getSubmitRecords()
        return json.loads(json.dumps(event))

    def setWorkInProgress(self, wip):
        # Gerrit only includes 'wip' in the data returned via ssh if
        # the value is true.
        if wip:
            self.data['wip'] = True
        elif 'wip' in self.data:
            del self.data['wip']

    def getSubmitRecords(self):
        status = {}
        for cat in self.categories:
            status[cat] = 0

        for a in self.patchsets[-1]['approvals']:
            cur = status[a['type']]
            cat_min, cat_max = self.categories[a['type']][1:]
            new = int(a['value'])
            if new == cat_min:
                cur = new
            elif abs(new) > abs(cur):
                cur = new
            status[a['type']] = cur

        labels = []
        ok = True
        for typ, cat in self.categories.items():
            cur = status[typ]
            cat_min, cat_max = cat[1:]
            if cur == cat_min:
                value = 'REJECT'
                ok = False
            elif cur == cat_max:
                value = 'OK'
            else:
                value = 'NEED'
                ok = False
            labels.append({'label': cat[0], 'status': value})
        if ok:
            return [{'status': 'OK'}]
        return [{'status': 'NOT_READY',
                 'labels': labels}]

    def getSubmitRequirements(self):
        return self.submit_requirements

    def setSubmitRequirements(self, reqs):
        self.submit_requirements = reqs

    def setDependsOn(self, other, patchset):
        self.depends_on_change = other
        self.depends_on_patchset = patchset
        d = {'id': other.data['id'],
             'number': other.data['number'],
             'ref': other.patchsets[patchset - 1]['ref']
             }
        self.data['dependsOn'] = [d]

        other.needed_by_changes.append((self, len(self.patchsets)))
        needed = other.data.get('neededBy', [])
        d = {'id': self.data['id'],
             'number': self.data['number'],
             'ref': self.patchsets[-1]['ref'],
             'revision': self.patchsets[-1]['revision']
             }
        needed.append(d)
        other.data['neededBy'] = needed

    def query(self):
        self.queried += 1
        d = self.data.get('dependsOn')
        if d:
            d = d[0]
            if (self.depends_on_change.patchsets[-1]['ref'] == d['ref']):
                d['isCurrentPatchSet'] = True
            else:
                d['isCurrentPatchSet'] = False
        return json.loads(json.dumps(self.data))

    def queryHTTP(self, internal=False):
        if not internal:
            self.queried += 1
        labels = {}
        for cat in self.categories:
            labels[cat] = {}
        for app in self.patchsets[-1]['approvals']:
            label = labels[app['type']]
            _, label_min, label_max = self.categories[app['type']]
            val = int(app['value'])
            label_all = label.setdefault('all', [])
            approval = {
                "value": val,
                "username": app['by']['username'],
                "email": app['by']['email'],
                "date": str(datetime.datetime.fromtimestamp(app['grantedOn'])),
            }
            if app.get('__tag') is not None:
                approval['tag'] = app['__tag']
            label_all.append(approval)
            if val == label_min:
                label['blocking'] = True
                if 'rejected' not in label:
                    label['rejected'] = app['by']
            if val == label_max:
                if 'approved' not in label:
                    label['approved'] = app['by']
        revisions = {}
        for i, rev in enumerate(self.patchsets):
            num = i + 1
            files = {}
            for f in rev['files']:
                if f['file'] == '/COMMIT_MSG':
                    continue
                files[f['file']] = {"status": f['type'][0]}  # ADDED -> A
            parent = '0000000000000000000000000000000000000000'
            if self.depends_on_change:
                parent = self.depends_on_change.patchsets[
                    self.depends_on_patchset - 1]['revision']
            parents = self.data.get('parents', [])
            if len(parents) <= 1:
                parents = [parent]
            revisions[rev['revision']] = {
                "kind": "REWORK",
                "_number": num,
                "created": rev['createdOn'],
                "uploader": rev['uploader'],
                "ref": rev['ref'],
                "commit": {
                    "subject": self.subject,
                    "message": self.data['commitMessage'],
                    "parents": [{
                        "commit": p,
                    } for p in parents]
                },
                "files": files
            }
        if self.cherry_pick:
            submit_type = 'CHERRY_PICK'
        else:
            submit_type = 'MERGE_IF_NECESSARY'
        data = {
            "id": self.project + '~' + self.branch + '~' + self.data['id'],
            "project": self.project,
            "branch": self.branch,
            "hashtags": [],
            "change_id": self.data['id'],
            "subject": self.subject,
            "status": self.data['status'],
            "created": self.data['createdOn'],
            "updated": self.data['lastUpdated'],
            "_number": self.number,
            "owner": self.data['owner'],
            "labels": labels,
            "current_revision": self.patchsets[-1]['revision'],
            "revisions": revisions,
            "requirements": [],
            "work_in_progresss": self.data.get('wip', False),
            "submit_type": submit_type,
        }
        if 'parents' in self.data:
            data['parents'] = self.data['parents']
        if 'topic' in self.data:
            data['topic'] = self.data['topic']
        data['submit_requirements'] = self.getSubmitRequirements()
        return json.loads(json.dumps(data))

    def queryRevisionHTTP(self, revision):
        for ps in self.patchsets:
            if ps['revision'] == revision:
                break
        else:
            return None
        changes = []
        if self.depends_on_change:
            changes.append({
                "commit": {
                    "commit": self.depends_on_change.patchsets[
                        self.depends_on_patchset - 1]['revision'],
                },
                "_change_number": self.depends_on_change.number,
                "_revision_number": self.depends_on_patchset,
                "_current_revision_number":
                self.depends_on_change.latest_patchset,
            })
        for (needed_by_change, needed_by_patchset) in self.needed_by_changes:
            changes.append({
                "commit": {
                    "commit": needed_by_change.patchsets[
                        needed_by_patchset - 1]['revision'],
                },
                "_change_number": needed_by_change.number,
                "_revision_number": needed_by_patchset,
                "_current_revision_number": needed_by_change.latest_patchset,
            })
        return {"changes": changes}

    def queryFilesHTTP(self, revision, parent):
        for rev in self.patchsets:
            if rev['revision'] == revision:
                break
        else:
            return None

        files = {}
        # We're querying the list of files different in this commit
        # compared to the parent (ie, the list of files that we're
        # merging into the branch, not the files in this commit).  If
        # this is a merge commit and the test has specified such a
        # files list, use that.
        if (parent and len(self.data.get('parents', []))
            and self.merge_parent_files):
            rev_files = self.merge_parent_files or []
        else:
            # Otherwise, the simple case of "files in this commit"
            rev_files = rev['files']
        for f in rev_files:
            files[f['file']] = {"status": f['type'][0]}  # ADDED -> A
        return files

    def setMerged(self):
        if (self.depends_on_change and
                self.depends_on_change.data['status'] != 'MERGED'):
            return
        if self.fail_merge:
            return
        self.data['status'] = 'MERGED'
        self.data['open'] = False
        self.open = False

        path = os.path.join(self.upstream_root, self.project)
        repo = git.Repo(path)

        repo.head.reference = self.branch
        repo.head.reset(working_tree=True)
        if self.cherry_pick:
            repo.git.cherry_pick(self.patchsets[-1]['ref'], allow_empty=True)
        else:
            repo.git.merge('-s', 'resolve', self.patchsets[-1]['ref'])
        repo.heads[self.branch].commit = repo.head.commit
        if self.cherry_pick:
            self.addCherryPickPatchset()

    def setReported(self):
        self.reported += 1

    def setAbandoned(self):
        self.data['status'] = 'ABANDONED'
        self.data['open'] = False
        self.open = False


class FakeGerritPoller(gerritconnection.GerritChecksPoller):
    """A Fake Gerrit poller for use in tests.

    This subclasses
    :py:class:`~zuul.connection.gerrit.GerritPoller`.
    """

    poll_interval = 1

    def _poll(self, *args, **kw):
        r = super(FakeGerritPoller, self)._poll(*args, **kw)
        # Set the event so tests can confirm that the poller has run
        # after they changed something.
        self.connection._poller_event.set()
        return r


class FakeGerritRefWatcher(gitwatcher.GitWatcher):
    """A Fake Gerrit ref watcher.

    This subclasses
    :py:class:`~zuul.connection.git.GitWatcher`.
    """

    def __init__(self, *args, **kw):
        super(FakeGerritRefWatcher, self).__init__(*args, **kw)
        self.baseurl = self.connection.upstream_root
        self.poll_delay = 1

    def _poll(self, *args, **kw):
        r = super(FakeGerritRefWatcher, self)._poll(*args, **kw)
        # Set the event so tests can confirm that the watcher has run
        # after they changed something.
        self.connection._ref_watcher_event.set()
        return r


class GerritWebServer(object):

    def __init__(self, fake_gerrit):
        super(GerritWebServer, self).__init__()
        self.fake_gerrit = fake_gerrit

    def start(self):
        fake_gerrit = self.fake_gerrit

        class Server(http.server.SimpleHTTPRequestHandler):
            log = logging.getLogger("zuul.test.FakeGerritConnection")
            review_re = re.compile('/a/changes/(.*?)/revisions/(.*?)/review')
            together_re = re.compile('/a/changes/(.*?)/submitted_together')
            submit_re = re.compile('/a/changes/(.*?)/submit')
            pending_checks_re = re.compile(
                r'/a/plugins/checks/checks\.pending/\?'
                r'query=checker:(.*?)\+\(state:(.*?)\)')
            update_checks_re = re.compile(
                r'/a/changes/(.*)/revisions/(.*?)/checks/(.*)')
            list_checkers_re = re.compile('/a/plugins/checks/checkers/')
            change_re = re.compile(r'/a/changes/(.*)\?o=.*')
            related_re = re.compile(r'/a/changes/(.*)/revisions/(.*)/related')
            files_parent_re = re.compile(
                r'/a/changes/(.*)/revisions/(.*)/files'
                r'\?parent=1')
            files_re = re.compile(r'/a/changes/(.*)/revisions/(.*)/files')
            change_search_re = re.compile(r'/a/changes/\?n=500.*&q=(.*)')
            version_re = re.compile(r'/a/config/server/version')
            info_re = re.compile(r'/a/config/server/info')
            head_re = re.compile(r'/a/projects/(.*)/HEAD')

            def do_POST(self):
                path = self.path
                self.log.debug("Got POST %s", path)
                fake_gerrit.api_calls.append(('POST', path))

                data = self.rfile.read(int(self.headers['Content-Length']))
                data = json.loads(data.decode('utf-8'))
                self.log.debug("Got data %s", data)

                m = self.review_re.match(path)
                if m:
                    return self.review(m.group(1), m.group(2), data)
                m = self.submit_re.match(path)
                if m:
                    return self.submit(m.group(1), data)
                m = self.update_checks_re.match(path)
                if m:
                    return self.update_checks(
                        m.group(1), m.group(2), m.group(3), data)
                self.send_response(500)
                self.end_headers()

            def do_GET(self):
                path = self.path
                self.log.debug("Got GET %s", path)
                fake_gerrit.api_calls.append(('GET', path))
                if fake_gerrit._fake_return_api_error:
                    self.send_response(500)
                    self.end_headers()
                    return

                m = self.change_re.match(path)
                if m:
                    return self.get_change(m.group(1))
                m = self.related_re.match(path)
                if m:
                    return self.get_related(m.group(1), m.group(2))
                m = self.files_parent_re.match(path)
                if m:
                    return self.get_files(m.group(1), m.group(2), 1)
                m = self.files_re.match(path)
                if m:
                    return self.get_files(m.group(1), m.group(2))
                m = self.together_re.match(path)
                if m:
                    return self.get_submitted_together(m.group(1))
                m = self.change_search_re.match(path)
                if m:
                    return self.get_changes(m.group(1))
                m = self.pending_checks_re.match(path)
                if m:
                    return self.get_pending_checks(m.group(1), m.group(2))
                m = self.list_checkers_re.match(path)
                if m:
                    return self.list_checkers()
                m = self.version_re.match(path)
                if m:
                    return self.version()
                m = self.info_re.match(path)
                if m:
                    return self.info()
                m = self.head_re.match(path)
                if m:
                    return self.head(m.group(1))
                self.send_response(500)
                self.end_headers()

            def _403(self, msg):
                self.send_response(403)
                self.end_headers()
                self.wfile.write(msg.encode('utf8'))

            def _400(self, msg):
                self.send_response(400)
                self.end_headers()
                self.wfile.write(msg.encode('utf8'))

            def _404(self):
                self.send_response(404)
                self.end_headers()

            def _409(self):
                self.send_response(409)
                self.end_headers()

            def _get_change(self, change_id):
                change_id = urllib.parse.unquote(change_id)
                project, branch, change = change_id.split('~')
                for c in fake_gerrit.changes.values():
                    if (c.data['id'] == change and
                        c.data['branch'] == branch and
                        c.data['project'] == project):
                        return c

            def review(self, change_id, revision, data):
                change = self._get_change(change_id)
                if not change:
                    return self._404()

                message = data['message']
                b_len = len(message.encode('utf-8'))
                if b_len > gerritreporter.GERRIT_HUMAN_MESSAGE_LIMIT:
                    self.send_response(400, message='Message length exceeded')
                    self.end_headers()
                    return
                labels = data.get('labels', {})
                comments = data.get('robot_comments', data.get('comments', {}))
                tag = data.get('tag', None)
                try:
                    fake_gerrit._test_handle_review(
                        int(change.data['number']), message, False, labels,
                        None, True, False, comments, tag=tag)
                except Exception as e:
                    return self._400(str(e))
                self.send_response(200)
                self.end_headers()

            def submit(self, change_id, data):
                change = self._get_change(change_id)
                if not change:
                    return self._404()

                if not fake_gerrit._fake_submit_permission:
                    return self._403('submit not permitted')

                candidate = self._get_change(change_id)
                sr = candidate.getSubmitRecords()
                if sr[0]['status'] != 'OK':
                    # One of the changes in this topic isn't
                    # ready to merge
                    return self._409()
                changes_to_merge = set(change.data['number'])
                if fake_gerrit._fake_submit_whole_topic:
                    results = fake_gerrit._test_get_submitted_together(change)
                    for record in results:
                        candidate = self._get_change(record['id'])
                        sr = candidate.getSubmitRecords()
                        if sr[0]['status'] != 'OK':
                            # One of the changes in this topic isn't
                            # ready to merge
                            return self._409()
                        changes_to_merge.add(candidate.data['number'])
                message = None
                labels = {}
                for change_number in changes_to_merge:
                    fake_gerrit._test_handle_review(
                        int(change_number), message, True, labels,
                        None, False, True)
                self.send_response(200)
                self.end_headers()

            def update_checks(self, change_id, revision, checker, data):
                self.log.debug("Update checks %s %s %s",
                               change_id, revision, checker)
                change = self._get_change(change_id)
                if not change:
                    return self._404()

                change.setCheck(checker, **data)
                self.send_response(200)
                # TODO: return the real data structure, but zuul
                # ignores this now.
                self.end_headers()

            def get_pending_checks(self, checker, state):
                self.log.debug("Get pending checks %s %s", checker, state)
                ret = []
                for c in fake_gerrit.changes.values():
                    if checker not in c.checks:
                        continue
                    patchset_pending_checks = {}
                    if c.checks[checker]['state'] == state:
                        patchset_pending_checks[checker] = {
                            'state': c.checks[checker]['state'],
                        }
                    if patchset_pending_checks:
                        ret.append({
                            'patch_set': {
                                'repository': c.project,
                                'change_number': c.number,
                                'patch_set_id': c.latest_patchset,
                            },
                            'pending_checks': patchset_pending_checks,
                        })
                self.send_data(ret)

            def list_checkers(self):
                self.log.debug("Get checkers")
                self.send_data(fake_gerrit.fake_checkers)

            def get_change(self, number):
                change = fake_gerrit.changes.get(int(number))
                if not change:
                    return self._404()

                self.send_data(change.queryHTTP())
                self.end_headers()

            def get_related(self, number, revision):
                change = fake_gerrit.changes.get(int(number))
                if not change:
                    return self._404()
                data = change.queryRevisionHTTP(revision)
                if data is None:
                    return self._404()
                self.send_data(data)
                self.end_headers()

            def get_files(self, number, revision, parent=None):
                change = fake_gerrit.changes.get(int(number))
                if not change:
                    return self._404()
                data = change.queryFilesHTTP(revision, parent)
                if data is None:
                    return self._404()
                self.send_data(data)
                self.end_headers()

            def get_submitted_together(self, number):
                change = fake_gerrit.changes.get(int(number))
                if not change:
                    return self._404()

                results = fake_gerrit._test_get_submitted_together(change)
                self.send_data(results)
                self.end_headers()

            def get_changes(self, query):
                self.log.debug("simpleQueryHTTP: %s", query)
                query = urllib.parse.unquote(query)
                fake_gerrit.queries.append(query)
                results = []
                if query.startswith('(') and 'OR' in query:
                    query = query[1:-1]
                    for q in query.split(' OR '):
                        for r in fake_gerrit._simpleQuery(q, http=True):
                            if r not in results:
                                results.append(r)
                else:
                    results = fake_gerrit._simpleQuery(query, http=True)
                self.send_data(results)
                self.end_headers()

            def version(self):
                self.send_data('3.0.0-some-stuff')
                self.end_headers()

            def info(self):
                # This is not complete; see documentation for
                # additional fields.
                data = {
                    "accounts": {
                        "visibility": "ALL",
                        "default_display_name": "FULL_NAME"
                    },
                    "change": {
                        "allow_blame": True,
                        "disable_private_changes": True,
                        "update_delay": 300,
                        "mergeability_computation_behavior":
                        "API_REF_UPDATED_AND_CHANGE_REINDEX",
                        "enable_robot_comments": True,
                        "conflicts_predicate_enabled": True
                    },
                    "gerrit": {
                        "all_projects": "All-Projects",
                        "all_users": "All-Users",
                        "doc_search": True,
                    },
                    "note_db_enabled": True,
                    "sshd": {},
                    "suggest": {"from": 0},
                    "user": {"anonymous_coward_name": "Name of user not set"},
                    "receive": {"enable_signed_push": False},
                    "submit_requirement_dashboard_columns": []
                }
                if fake_gerrit._fake_submit_whole_topic:
                    data['change']['submit_whole_topic'] = True
                self.send_data(data)
                self.end_headers()

            def head(self, project):
                project = urllib.parse.unquote(project)
                head = fake_gerrit._fake_project_default_branch.get(
                    project, 'master')
                self.send_data('refs/heads/' + head)
                self.end_headers()

            def send_data(self, data):
                data = json.dumps(data).encode('utf-8')
                data = b")]}'\n" + data
                self.send_response(200)
                self.send_header('Content-Type', 'application/json')
                self.send_header('Content-Length', len(data))
                self.end_headers()
                self.wfile.write(data)

            def log_message(self, fmt, *args):
                self.log.debug(fmt, *args)

        self.httpd = socketserver.ThreadingTCPServer(('', 0), Server)
        self.port = self.httpd.socket.getsockname()[1]
        self.thread = threading.Thread(name='GerritWebServer',
                                       target=self.httpd.serve_forever)
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        self.httpd.shutdown()
        self.thread.join()
        self.httpd.server_close()


class FakeGerritConnection(gerritconnection.GerritConnection):
    """A Fake Gerrit connection for use in tests.

    This subclasses
    :py:class:`~zuul.connection.gerrit.GerritConnection` to add the
    ability for tests to add changes to the fake Gerrit it represents.
    """

    log = logging.getLogger("zuul.test.FakeGerritConnection")
    _poller_class = FakeGerritPoller
    _ref_watcher_class = FakeGerritRefWatcher

    def __init__(self, driver, connection_name, connection_config,
                 changes_db=None, upstream_root=None, poller_event=None,
                 ref_watcher_event=None, submit_whole_topic=None):

        if connection_config.get('password'):
            self.web_server = GerritWebServer(self)
            self.web_server.start()
            url = 'http://localhost:%s' % self.web_server.port
            connection_config['baseurl'] = url
        else:
            self.web_server = None

        super(FakeGerritConnection, self).__init__(driver, connection_name,
                                                   connection_config)

        self.fixture_dir = os.path.join(FIXTURE_DIR, 'gerrit')
        self.change_number = 0
        self.changes = changes_db
        self.queries = []
        self.api_calls = []
        self._fake_return_api_error = False
        self.upstream_root = upstream_root
        self.fake_checkers = []
        self._poller_event = poller_event
        self._ref_watcher_event = ref_watcher_event
        self._fake_submit_whole_topic = bool(submit_whole_topic)
        self._fake_submit_permission = True
        self._fake_project_default_branch = {}
        self.submit_retry_backoff = 0

    def onStop(self):
        super().onStop()
        if self.web_server:
            self.web_server.stop()

    def addFakeChecker(self, **kw):
        self.fake_checkers.append(kw)

    def addFakeChange(self, project, branch, subject, status='NEW',
                      files=None, parent=None, merge_parents=None,
                      merge_files=None, topic=None, empty=False):
        """Add a change to the fake Gerrit."""
        self.change_number += 1
        c = FakeGerritChange(self, self.change_number, project, branch,
                             subject, upstream_root=self.upstream_root,
                             status=status, files=files, parent=parent,
                             merge_parents=merge_parents,
                             merge_files=merge_files,
                             topic=topic, empty=empty)
        self.changes[self.change_number] = c
        return c

    def addFakeTag(self, project, branch, tag, message=None):
        path = os.path.join(self.upstream_root, project)
        repo = git.Repo(path)
        commit = repo.heads[branch].commit
        ref = 'refs/tags/' + tag

        t = git.Tag.create(repo, tag, commit, logmsg=message)
        newrev = t.object.hexsha

        event = {
            "type": "ref-updated",
            "submitter": {
                "name": "User Name",
            },
            "refUpdate": {
                "oldRev": 40 * '0',
                "newRev": newrev,
                "refName": ref,
                "project": project,
            }
        }
        return event

    def getFakeBranchCreatedEvent(self, project, branch):
        path = os.path.join(self.upstream_root, project)
        repo = git.Repo(path)
        oldrev = 40 * '0'

        event = {
            "type": "ref-updated",
            "submitter": {
                "name": "User Name",
            },
            "refUpdate": {
                "oldRev": oldrev,
                "newRev": repo.heads[branch].commit.hexsha,
                "refName": 'refs/heads/' + branch,
                "project": project,
            }
        }
        return event

    def getFakeBranchDeletedEvent(self, project, branch):
        oldrev = '4abd38457c2da2a72d4d030219ab180ecdb04bf0'
        newrev = 40 * '0'

        event = {
            "type": "ref-updated",
            "submitter": {
                "name": "User Name",
            },
            "refUpdate": {
                "oldRev": oldrev,
                "newRev": newrev,
                "refName": 'refs/heads/' + branch,
                "project": project,
            }
        }
        return event

    def getProjectHeadUpdatedEvent(self, project, old, new):
        event = {
            "projectName": project,
            "oldHead": f"refs/heads/{old}",
            "newHead": f"refs/heads/{new}",
            "type": "project-head-updated",
        }
        return event

    def review(self, item, change, message, submit, labels,
               checks_api, notify, file_comments, phase1, phase2,
               zuul_event_id=None):
        if self.web_server:
            return super(FakeGerritConnection, self).review(
                item, change, message, submit, labels, checks_api,
                notify, file_comments, phase1, phase2,
                zuul_event_id)
        self._test_handle_review(int(change.number), message, submit,
                                 labels, notify, phase1, phase2)

    def _test_get_submitted_together(self, change):
        topic = change.data.get('topic')
        if not self._fake_submit_whole_topic:
            topic = None
        if topic:
            results = self._simpleQuery(f'status:open topic:{topic}',
                                        http=True)
        else:
            results = [change.queryHTTP(internal=True)]
        for dep in change.data.get('dependsOn', []):
            dep_change = self.changes.get(int(dep['number']))
            r = dep_change.queryHTTP(internal=True)
            if r['status'] == 'MERGED':
                continue
            if not dep_change.open:
                continue
            if r not in results:
                results.append(r)
        if len(results) == 1:
            return []
        return results

    def _test_handle_review(self, change_number, message, submit, labels,
                            notify,
                            phase1, phase2, file_comments=None, tag=None):
        # Handle a review action from a test
        change = self.changes[change_number]

        change.notify = notify

        # Add the approval back onto the change (ie simulate what gerrit would
        # do).
        # Usually when zuul leaves a review it'll create a feedback loop where
        # zuul's review enters another gerrit event (which is then picked up by
        # zuul). However, we can't mimic this behaviour (by adding this
        # approval event into the queue) as it stops jobs from checking what
        # happens before this event is triggered. If a job needs to see what
        # happens they can add their own verified event into the queue.
        # Nevertheless, we can update change with the new review in gerrit.

        if phase1:
            for cat in labels:
                change.addApproval(cat, labels[cat], username=self.user,
                                   tag=tag)

            if message:
                change.messages.append(message)

            if file_comments:
                ok_files = set(x['file']
                               for x in change.patchsets[-1]['files'])
                ok_files.add('/COMMIT_MSG')
                for filename, commentlist in file_comments.items():
                    if filename not in ok_files:
                        raise Exception(
                            f"file {filename} not found in revision")
                    for comment in commentlist:
                        change.addComment(filename, comment['line'],
                                          comment['message'], 'Zuul',
                                          'zuul@example.com', self.user,
                                          comment.get('range'))
            if message:
                change.setReported()
        if submit and phase2:
            change.setMerged()

    def queryChangeSSH(self, number, event=None):
        self.log.debug("Query change SSH: %s", number)
        change = self.changes.get(int(number))
        if change:
            return change.query()
        return {}

    def _simpleQuery(self, query, http=False):
        if http:
            def queryMethod(change):
                return change.queryHTTP()
        else:
            def queryMethod(change):
                return change.query()
        # the query can be in parenthesis so strip them if needed
        if query.startswith('('):
            query = query[1:-1]
        if query.startswith('change:'):
            # Query a specific changeid
            changeid = query[len('change:'):]
            l = [queryMethod(change) for change in self.changes.values()
                 if (change.data['id'] == changeid or
                     change.data['number'] == changeid)]
        elif query.startswith('message:'):
            # Query the content of a commit message
            msg = query[len('message:'):].strip()
            # Remove quoting if it is there
            if msg.startswith('{') and msg.endswith('}'):
                msg = msg[1:-1]
            l = [queryMethod(change) for change in self.changes.values()
                 if msg in change.data['commitMessage']]
        else:
            cut_off_time = 0
            l = list(self.changes.values())
            parts = query.split(" ")
            for part in parts:
                if part.startswith("-age"):
                    _, _, age = part.partition(":")
                    cut_off_time = (
                        datetime.datetime.now().timestamp() - float(age[:-1])
                    )
                    l = [
                        change for change in l
                        if change.data["lastUpdated"] >= cut_off_time
                    ]
                if part.startswith('topic:'):
                    topic = part[len('topic:'):].strip().strip('"\'')
                    l = [
                        change for change in l
                        if 'topic' in change.data
                        and topic in change.data['topic']
                    ]
                if part.startswith('status:'):
                    status = part[len('status:'):].strip().strip('"\'').lower()
                    if status == 'open':
                        l = [
                            change for change in l if change.open
                        ]
                    else:
                        l = [
                            change for change in l
                            if change.data['status'].lower() == status
                        ]
            l = [queryMethod(change) for change in l]
        return l

    def simpleQuerySSH(self, query, event=None):
        log = get_annotated_logger(self.log, event)
        log.debug("simpleQuerySSH: %s", query)
        self.queries.append(query)
        results = []
        if query.startswith('(') and 'OR' in query:
            query = query[1:-1]
            for q in query.split(' OR '):
                for r in self._simpleQuery(q):
                    if r not in results:
                        results.append(r)
        else:
            results = self._simpleQuery(query)
        return results

    def startSSHListener(self, *args, **kw):
        pass

    def _uploadPack(self, project):
        ret = ('00a31270149696713ba7e06f1beb760f20d359c4abed HEAD\x00'
               'multi_ack thin-pack side-band side-band-64k ofs-delta '
               'shallow no-progress include-tag multi_ack_detailed no-done\n')
        path = os.path.join(self.upstream_root, project.name)
        repo = git.Repo(path)
        for ref in repo.refs:
            if ref.path.endswith('.lock'):
                # don't treat lockfiles as ref
                continue
            r = ref.object.hexsha + ' ' + ref.path + '\n'
            ret += '%04x%s' % (len(r) + 4, r)
        ret += '0000'
        return ret

    def getGitUrl(self, project):
        return 'file://' + os.path.join(self.upstream_root, project.name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import base64
import types

from zuul.lib import encryption

import yaml
from yaml import YAMLError  # noqa: F401


try:
    # Explicit type ignore to deal with provisional import failure
    # Details at https://github.com/python/mypy/issues/1153
    from yaml import cyaml
    import _yaml
    SafeLoader = cyaml.CSafeLoader
    SafeDumper = cyaml.CSafeDumper
    Mark = _yaml.Mark
except ImportError:
    SafeLoader = yaml.SafeLoader
    SafeDumper = yaml.SafeDumper
    Mark = yaml.Mark


class OverrideValue:
    def __init__(self, value, override):
        self.value = value
        self.override = override


# Generally it only makes sense to override lists or dicts, but
# because of the to_list construction, we might end up with strings
# too.
class OverrideStr(OverrideValue):
    pass


class OverrideList(OverrideValue):
    pass


class OverrideDict(OverrideValue):
    pass


class Override:
    yaml_tag = u'!override'
    override_value = True

    @classmethod
    def from_yaml(cls, loader, node):
        if isinstance(node, yaml.MappingNode):
            return OverrideDict(loader.construct_mapping(node),
                                cls.override_value)
        elif isinstance(node, yaml.SequenceNode):
            return OverrideList(loader.construct_sequence(node),
                                cls.override_value)
        elif isinstance(node, yaml.ScalarNode):
            tag = loader.resolve(yaml.ScalarNode, node.value, (True, False))
            node = yaml.ScalarNode(tag, node.value,
                                   node.start_mark, node.end_mark)
            raw_value = loader.construct_object(node)
            if isinstance(raw_value, str):
                return OverrideStr(raw_value, cls.override_value)
            raise Exception("Unsupported type for scalar override control: "
                            f"{type(raw_value)}")
        raise Exception(f"Unsupported type for override control: {type(node)}")


class Inherit(Override):
    yaml_tag = u'!inherit'
    override_value = False


class EncryptedPKCS1_OAEP:
    yaml_tag = u'!encrypted/pkcs1-oaep'

    def __init__(self, ciphertext):
        if isinstance(ciphertext, list):
            self.ciphertext = [base64.b64decode(x.value)
                               for x in ciphertext]
        else:
            self.ciphertext = base64.b64decode(ciphertext)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if not isinstance(other, EncryptedPKCS1_OAEP):
            return False
        return (self.ciphertext == other.ciphertext)

    @classmethod
    def from_yaml(cls, loader, node):
        return cls(node.value)

    @classmethod
    def to_yaml(cls, dumper, data):
        ciphertext = data.ciphertext
        if isinstance(ciphertext, list):
            ciphertext = [yaml.ScalarNode(tag='tag:yaml.org,2002:str',
                                          value=base64.b64encode(x))
                          for x in ciphertext]
            return yaml.SequenceNode(tag=cls.yaml_tag,
                                     value=ciphertext)
        ciphertext = base64.b64encode(ciphertext).decode('utf8')
        return yaml.ScalarNode(tag=cls.yaml_tag, value=ciphertext)

    def decrypt(self, private_key):
        if isinstance(self.ciphertext, list):
            return ''.join([
                encryption.decrypt_pkcs1_oaep(chunk, private_key).
                decode('utf8')
                for chunk in self.ciphertext])
        else:
            return encryption.decrypt_pkcs1_oaep(self.ciphertext,
                                                 private_key).decode('utf8')


class ZuulConfigKey(str):
    def __new__(cls, s, line):
        self = super().__new__(cls, s)
        self.line = line
        return self

    def __copy__(self):
        return ZuulConfigKey(self, self.line)

    def __deepcopy__(self, memo):
        return self.__copy__()

    @classmethod
    def to_yaml(cls, dumper, data):
        return yaml.representer.SafeRepresenter.represent_str(
            dumper, str(data))


def safe_load(stream, *args, **kwargs):
    return yaml.load(stream, *args, Loader=SafeLoader, **kwargs)


def safe_dump(stream, *args, **kwargs):
    return yaml.dump(stream, *args, Dumper=SafeDumper, **kwargs)


class EncryptedDumper(SafeDumper):
    pass


class EncryptedLoader(SafeLoader):
    pass


# Add support for encrypted objects
EncryptedDumper.add_representer(EncryptedPKCS1_OAEP,
                                EncryptedPKCS1_OAEP.to_yaml)
EncryptedLoader.add_constructor(EncryptedPKCS1_OAEP.yaml_tag,
                                EncryptedPKCS1_OAEP.from_yaml)
# Also add support for serializing frozen data
EncryptedDumper.add_representer(
    types.MappingProxyType,
    yaml.representer.SafeRepresenter.represent_dict)
EncryptedDumper.add_representer(
    ZuulConfigKey,
    ZuulConfigKey.to_yaml)
# Add support for override control
EncryptedLoader.add_constructor(Override.yaml_tag,
                                Override.from_yaml)
EncryptedLoader.add_constructor(Inherit.yaml_tag,
                                Inherit.from_yaml)


def encrypted_dump(data, *args, **kwargs):
    return yaml.dump(data, *args, Dumper=EncryptedDumper, **kwargs)


def encrypted_load(stream, *args, **kwargs):
    return yaml.load(stream, *args, Loader=EncryptedLoader, **kwargs)


# Add support for the Ansible !unsafe tag
# Note that "unsafe" here is used differently than "safe" from PyYAML

class AnsibleUnsafeStr:
    yaml_tag = u'!unsafe'

    def __init__(self, value):
        self.value = value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if isinstance(other, AnsibleUnsafeStr):
            return self.value == other.value
        return self.value == other

    @classmethod
    def from_yaml(cls, loader, node):
        return cls(node.value)

    @classmethod
    def to_yaml(cls, dumper, data):
        return yaml.ScalarNode(tag=cls.yaml_tag, value=data.value)


class AnsibleUnsafeDumper(SafeDumper):
    pass


class AnsibleUnsafeDumperWithoutAliases(SafeDumper):
    def ignore_aliases(self, data):
        return True


class AnsibleUnsafeLoader(SafeLoader):
    pass


AnsibleUnsafeDumper.add_representer(AnsibleUnsafeStr,
                                    AnsibleUnsafeStr.to_yaml)
AnsibleUnsafeLoader.add_constructor(AnsibleUnsafeStr.yaml_tag,
                                    AnsibleUnsafeStr.from_yaml)


def ansible_unsafe_dump(data, *args, **kwargs):
    ignore_aliases = kwargs.pop('ignore_aliases', False)
    if ignore_aliases:
        return yaml.dump(data, *args, Dumper=AnsibleUnsafeDumperWithoutAliases,
                         **kwargs)
    else:
        return yaml.dump(data, *args, Dumper=AnsibleUnsafeDumper, **kwargs)


def ansible_unsafe_load(stream, *args, **kwargs):
    return yaml.load(stream, *args, Loader=AnsibleUnsafeLoader, **kwargs)


def mark_strings_unsafe(d):
    """Traverse a json-style data structure and replace every string value
    with an AnsibleUnsafeStr

    Returns the new structure.
    """
    if isinstance(d, tuple):
        d = list(d)

    if isinstance(d, dict):
        newdict = {}
        for key, value in d.items():
            newdict[key] = mark_strings_unsafe(value)
        return newdict
    elif isinstance(d, list):
        return [mark_strings_unsafe(v) for v in d]
    elif isinstance(d, int):
        return d
    elif isinstance(d, float):
        return d
    elif isinstance(d, type(None)):
        return d
    elif isinstance(d, bool):
        return d
    elif isinstance(d, AnsibleUnsafeStr):
        return d
    elif isinstance(d, str):
        return AnsibleUnsafeStr(d)
    else:
        raise Exception("Unhandled type: %s", type(d))



# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import confluent_kafka as kafka
import logging
import pprint
import threading

# With multiple Kafka partitions, events could arrive out of order.
# Similar to webhooks, we accept that and don't do anything to
# mitigate that.  They should mostly arrive in the correct order, at
# least at human scale.  That is, while we are somewhat likely to see
# an update to a change and the corresponding update to refs/meta out
# of order, we are unlikely to see a change abandoned before it is
# created.


class GerritKafkaEventListener:
    log = logging.getLogger("zuul.GerritConnection.kafka")
    RECONNECTION_DELAY = 5

    def __init__(self, gerrit_connection, connection_config):
        self.gerrit_connection = gerrit_connection
        bs = connection_config.get('kafka_bootstrap_servers')
        kafka_config = {
            'bootstrap.servers': bs,
        }
        kafka_config['client.id'] = connection_config.get(
            'kafka_client_id', 'zuul')
        kafka_config['group.id'] = connection_config.get(
            'kafka_group_id', 'zuul')
        tls_key = connection_config.get('kafka_tls_key', None)
        tls_cert = connection_config.get('kafka_tls_cert', None)
        tls_ca = connection_config.get('kafka_tls_ca', None)
        if tls_key:
            kafka_config['ssl.key.location'] = tls_key
            kafka_config['ssl.certificate.location'] = tls_cert
            kafka_config['ssl.ca.location'] = tls_ca
        self.kafka_config = kafka_config
        self.topic = connection_config.get('kafka_topic', 'gerrit')
        self._stop_event = threading.Event()
        self._stopped = False
        self._thread = None

    def start(self):
        self._thread = threading.Thread(target=self.run)
        self._thread.start()

    def stop(self):
        self.log.debug("Stopping kafka listener")
        self._stopped = True
        self._stop_event.set()

    def join(self):
        if self._thread:
            self._thread.join()

    def _run(self):
        self.log.info("Connecting to kafka at %s",
                      self.kafka_config['bootstrap.servers'])
        consumer = kafka.Consumer(self.kafka_config, logger=self.log)
        # So the unit tests can access it
        self.consumer = consumer
        try:
            consumer.subscribe([self.topic])
            while not self._stopped:
                msg = consumer.poll(timeout=2.0)
                if msg is None:
                    continue

                if msg.error():
                    if msg.error().code() == kafka.KafkaError._PARTITION_EOF:
                        self.log.info(
                            "Kafka topic %s partition %s "
                            "reached end at offset %s",
                            msg.topic(), msg.partition(), msg.offset())
                    else:
                        raise kafka.KafkaException(msg.error())
                else:
                    data = json.loads(msg.value().decode('utf8'))
                    self.log.info("Received data from kafka: \n%s" %
                                  pprint.pformat(data))
                    self.gerrit_connection.addEvent(data)
        finally:
            consumer.close()

    def run(self):
        while not self._stopped:
            try:
                self._run()
            except Exception:
                self.log.exception("Exception in kafka consumer with %s:",
                                   self.gerrit_connection.connection_name)
                self._stop_event.wait(self.RECONNECTION_DELAY)



# Copyright 2025 Acme Gating, LLC
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""system_event_table

Revision ID: ce1459953a12
Revises: 6c1582c1d08c
Create Date: 2025-11-19 11:18:16.472373

"""

# revision identifiers, used by Alembic.
revision = 'ce1459953a12'
down_revision = '6c1582c1d08c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

SYSTEM_EVENT_TABLE = "zuul_system_event"


def upgrade(table_prefix=''):
    op.create_table(
        table_prefix + SYSTEM_EVENT_TABLE,
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column('tenant', sa.String(255)),
        sa.Column('event_id', sa.String(255)),
        sa.Column("event_time", sa.DateTime),
        sa.Column("event_type", sa.String(255)),
        sa.Column("description", sa.TEXT()),
    )
    op.create_index(
        f'{table_prefix}zuul_system_event_tenant_idx',
        f'{table_prefix}zuul_system_event', ['tenant'])
    op.create_index(
        f'{table_prefix}zuul_system_event_event_id_idx',
        f'{table_prefix}zuul_system_event', ['event_id'])
    op.create_index(
        f'{table_prefix}zuul_system_event_event_time_idx',
        f'{table_prefix}zuul_system_event', ['event_time'])
    op.create_index(
        f'{table_prefix}zuul_system_event_event_type_idx',
        f'{table_prefix}zuul_system_event', ['event_type'])


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2013 OpenStack Foundation
# Copyright 2021-2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import argparse
import configparser
import daemon
import extras
import io
import json
import logging
import logging.config
import os
import signal
import socket
import sys
import traceback
import threading

yappi = extras.try_import('yappi')
objgraph = extras.try_import('objgraph')

# as of python-daemon 1.6 it doesn't bundle pidlockfile anymore
# instead it depends on lockfile-0.9.1 which uses pidfile.
pid_file_module = extras.try_imports(['daemon.pidlockfile', 'daemon.pidfile'])

from zuul.ansible import logconfig
import zuul.lib.connections
from zuul.lib.config import get_default


def stack_dump_handler(signum, frame):
    signal.signal(signal.SIGUSR2, signal.SIG_IGN)
    log = logging.getLogger("zuul.stack_dump")
    log.debug("Beginning debug handler")
    try:
        if yappi:
            if yappi.is_running():
                log.debug("Stopping Yappi")
                yappi.stop()
                yappi_out = io.StringIO()
                yappi.get_func_stats().print_all(out=yappi_out)
                yappi.get_thread_stats().print_all(out=yappi_out)
                log.debug(yappi_out.getvalue())
                yappi_out.close()
                yappi.clear_stats()
    except Exception:
        log.exception("Yappi error:")
    try:
        threads = {}
        for t in threading.enumerate():
            threads[t.ident] = t
        log_str = ""
        for thread_id, stack_frame in sys._current_frames().items():
            thread = threads.get(thread_id)
            if thread:
                thread_name = thread.name
                thread_is_daemon = str(thread.daemon)
            else:
                thread_name = '(Unknown)'
                thread_is_daemon = '(Unknown)'
            log_str += "Thread: %s %s d: %s\n"\
                       % (thread_id, thread_name, thread_is_daemon)
            log_str += "".join(traceback.format_stack(stack_frame))
        log.debug(log_str)
    except Exception:
        log.exception("Thread dump error:")
    try:
        if objgraph:
            log.debug("Most common types:")
            objgraph_out = io.StringIO()
            objgraph.show_growth(limit=100, file=objgraph_out)
            log.debug(objgraph_out.getvalue())
            objgraph_out.close()
    except Exception:
        log.exception("Objgraph error:")
    try:
        if yappi:
            if not yappi.is_running():
                log.debug("Starting Yappi")
                yappi.start()
    except Exception:
        log.exception("Yappi error:")
    log.debug("End debug handler")
    signal.signal(signal.SIGUSR2, stack_dump_handler)


class ZuulApp(object):
    app_name = None  # type: str
    app_description = None  # type: str

    def __init__(self):
        self.args = None
        self.config = None
        self.connections = {}
        self.commands = {}

    def _get_version(self):
        from zuul.version import release_string
        return "Zuul version: %s" % release_string

    def createParser(self):
        parser = argparse.ArgumentParser(
            description=self.app_description,
            formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument('-c', dest='config',
                            help='specify the config file')
        parser.add_argument('--version', dest='version', action='version',
                            version=self._get_version(),
                            help='show zuul version')
        return parser

    def addSubCommands(self, parser, commands):
        # Add a list of commandsocket.Command items to the parser
        subparsers = parser.add_subparsers(
            title='Online commands',
            description=('The following commands may be used to affect '
                         'the running process.'),
        )
        for command in commands:
            self.commands[command.name] = command
            cmd = subparsers.add_parser(
                command.name, help=command.help)
            cmd.set_defaults(command=command.name)
            for arg in command.args:
                cmd.add_argument(arg.name,
                                 help=arg.help,
                                 default=arg.default)

    def handleCommands(self):
        command_name = getattr(self.args, 'command', None)
        if command_name in self.commands:
            command = self.commands[self.args.command]
            command_str = command.name
            command_args = [getattr(self.args, arg.name)
                            for arg in command.args]
            if command_args:
                command_str += ' ' + json.dumps(command_args)
            self.sendCommand(command_str)
            sys.exit(0)

    def parseArguments(self, args=None):
        parser = self.createParser()
        self.args = parser.parse_args(args)

        if getattr(self.args, 'foreground', False):
            self.args.nodaemon = True
        else:
            self.args.nodaemon = False

        if getattr(self.args, 'command', None):
            self.args.nodaemon = True

        return parser

    def envBool(self, name, default=None):
        """Get the named variable from the environment and
        convert to boolean"""
        val = os.getenv(name)
        if val is None:
            return default
        if val.lower() in ['false', '0']:
            return False
        return True

    def readConfig(self):
        safe_env = {
            k: v for k, v in os.environ.items()
            if k.startswith('ZUUL_')
        }
        self.config = configparser.ConfigParser(safe_env)
        if self.args.config:
            locations = [self.args.config]
        else:
            locations = ['/etc/zuul/zuul.conf',
                         '~/zuul.conf']
        for fp in locations:
            if os.path.exists(os.path.expanduser(fp)):
                self.config.read(os.path.expanduser(fp))
                return
        raise Exception("Unable to locate config file in %s" % locations)

    def setup_logging(self, section, parameter):
        if self.config.has_option(section, parameter):
            fp = os.path.expanduser(self.config.get(section, parameter))
            logging_config = logconfig.load_config(fp)
        else:
            # If someone runs in the foreground and doesn't give a logging
            # config, leave the config set to emit to stdout.
            if hasattr(self.args, 'nodaemon') and self.args.nodaemon:
                logging_config = logconfig.ServerLoggingConfig()
            else:
                # Setting a server value updates the defaults to use
                # WatchedFileHandler on /var/log/zuul/{server}-debug.log
                # and /var/log/zuul/{server}.log
                logging_config = logconfig.ServerLoggingConfig(server=section)
            if hasattr(self.args, 'debug') and self.args.debug:
                logging_config.setDebug()
        logging_config.apply()

    def configure_connections(self, database=False, sources=False,
                              triggers=False, reporters=False,
                              providers=False, check_bwrap=False):
        self.connections = zuul.lib.connections.ConnectionRegistry(
            check_bwrap=check_bwrap)
        self.connections.configure(self.config, database=database,
                                   sources=sources, triggers=triggers,
                                   reporters=reporters, providers=providers)


class ZuulDaemonApp(ZuulApp, metaclass=abc.ABCMeta):
    def createParser(self):
        parser = super(ZuulDaemonApp, self).createParser()
        parser.add_argument('-d', dest='debug', action='store_true',
                            help='enable debug log')
        parser.add_argument('-f', dest='foreground', action='store_true',
                            help='run in foreground')
        return parser

    def getPidFile(self):
        pid_fn = get_default(self.config, self.app_name, 'pidfile',
                             '/var/run/zuul/%s.pid' % self.app_name,
                             expand_user=True)
        return pid_fn

    @abc.abstractmethod
    def run(self):
        """
        This is the main run method of the application.
        """
        pass

    def setup_logging(self, section, parameter):
        super(ZuulDaemonApp, self).setup_logging(section, parameter)
        from zuul.version import release_string
        log = logging.getLogger(
            "zuul.{section}".format(section=section.title()))
        log.debug(
            "Configured logging: {version}".format(
                version=release_string))

    def main(self):
        self.parseArguments()
        self.readConfig()

        pid_fn = self.getPidFile()
        pid = pid_file_module.TimeoutPIDLockFile(pid_fn, 10)

        # Early register the stack dump handler for all zuul apps. This makes
        # it possible to also gather stack dumps during startup hangs.
        signal.signal(signal.SIGUSR2, stack_dump_handler)

        if self.args.nodaemon:
            self.run()
        else:
            # Exercise the pidfile before we do anything else (including
            # logging or daemonizing)
            with pid:
                pass

            with daemon.DaemonContext(pidfile=pid, umask=0o022):
                self.run()

    def sendCommand(self, cmd):
        command_socket = get_default(
            self.config, self.app_name, 'command_socket',
            '/var/lib/zuul/%s.socket' % self.app_name)
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        s.connect(command_socket)
        cmd = '%s\n' % cmd
        s.sendall(cmd.encode('utf8'))



# Copyright 2021-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import time
import os
import re
import urllib
import uuid

import fixtures

import responses
import requests


class CRUDManager:
    name = ''

    def __init__(self, cloud):
        self.cloud = cloud
        self.items = []
        self.requests = []

    def list(self, request):
        resp = {'value': self.items}
        return (200, {}, json.dumps(resp))

    def get(self, request):
        url = urllib.parse.urlparse(request.path_url)
        for item in self.items:
            if item['id'] == url.path:
                return (200, {}, json.dumps(item))
        return (404, {}, json.dumps({
            'error': {
                'message': 'Not Found',
                'code': 'NotFound',
            }}))


class ResourceGroupsCRUD(CRUDManager):
    name = "resourcegroups"

    def put(self, request):
        data = json.loads(request.body)
        url = urllib.parse.urlparse(request.path_url)
        name = url.path.split('/')[-1]
        data['id'] = url.path
        data['name'] = name
        data['type'] = "Microsoft.Resources/resourceGroups"
        data['provisioningState'] = 'Succeeded'

        self.items.append(data)
        return (200, {}, json.dumps(data))


class PublicIPAddressesCRUD(CRUDManager):
    name = "Microsoft.Network/publicIPAddresses"

    def put(self, request):
        return self._put(json.loads(request.body),
                         request.path_url)

    def _put(self, data, url):
        url = urllib.parse.urlparse(url)
        name = url.path.split('/')[-1]
        data['id'] = url.path
        data['name'] = name
        data['type'] = self.name
        data['properties'] = {
            "provisioningState": "Updating",
            "resourceGuid": str(uuid.uuid4()),
            "publicIPAddressVersion": "IPv4",
            "publicIPAllocationMethod": "Static",
            "idleTimeoutInMinutes": 4,
            "ipTags": [],
        }
        self.items.append(data)
        ret = json.dumps(data)
        # Finish provisioning after return
        data['properties']['ipAddress'] = "198.51.100.1"
        data['properties']['provisioningState'] = "Succeeded"
        return (200, {}, ret)

    def delete(self, request):
        url = urllib.parse.urlparse(request.path_url)
        name = url.path.split('/')[-1]
        for item in self.items:
            if item['name'] == name:
                self.items.remove(item)
                return (200, {}, '')
        return (404, {}, json.dumps({
            'error': {
                'message': 'Not Found',
                'code': 'NotFound',
            }}))


class NetworkInterfacesCRUD(CRUDManager):
    name = "Microsoft.Network/networkInterfaces"

    def put(self, request):
        return self._put(json.loads(request.body),
                         request.path_url)

    def _put(self, data, url):
        url = urllib.parse.urlparse(url)
        name = url.path.split('/')[-1]
        data['id'] = url.path
        data['name'] = name
        data['type'] = self.name
        ipconfig = data['properties']['ipConfigurations'][0]
        data['properties'] = {
            "provisioningState": "Succeeded",
            "resourceGuid": str(uuid.uuid4()),
            "ipConfigurations": [
                {
                    "name": ipconfig['name'],
                    "id": os.path.join(data['id'], ipconfig['name']),
                    "type": ("Microsoft.Network/networkInterfaces/"
                             "ipConfigurations"),
                    "properties": {
                        "provisioningState": "Succeeded",
                        "privateIPAddress": "10.0.0.4",
                        "privateIPAllocationMethod": "Dynamic",
                        # publicIPAddress conditionally added below
                        "subnet": ipconfig['properties']['subnet'],
                        "primary": True,
                        "privateIPAddressVersion": "IPv4",
                    },
                }
            ],
            "enableAcceleratedNetworking": False,
            "enableIPForwarding": False,
            "hostedWorkloads": [],
            "tapConfigurations": [],
            "nicType": "Standard"
        }
        if 'publicIPAddress' in ipconfig['properties']:
            data['properties']['ipConfigurations'][0]['properties'][
                'publicIPAddress'] = ipconfig['properties']['publicIPAddress']
        self.items.append(data)
        return (200, {}, json.dumps(data))

    def delete(self, request):
        url = urllib.parse.urlparse(request.path_url)
        name = url.path.split('/')[-1]
        for item in self.items:
            if item['name'] == name:
                self.items.remove(item)
                return (200, {}, '')
        return (404, {}, json.dumps({
            'error': {
                'message': 'Not Found',
                'code': 'NotFound',
            }}))


class VirtualMachinesCRUD(CRUDManager):
    name = "Microsoft.Compute/virtualMachines"

    def put(self, request):
        data = json.loads(request.body)
        origdata = copy.deepcopy(data)
        self.requests.append(origdata)
        url = urllib.parse.urlparse(request.path_url)
        name = url.path.split('/')[-1]
        data['id'] = url.path
        data['name'] = name
        data['type'] = self.name
        data['properties'] = {
            "vmId": str(uuid.uuid4()),
            "hardwareProfile": data['properties']['hardwareProfile'],
            "storageProfile": {
                "imageReference": (data['properties']['storageProfile']
                                   ['imageReference']),
                "osDisk": {
                    "osType": "Linux",
                    "createOption": "FromImage",
                    "caching": "ReadWrite",
                    "managedDisk": {
                        "storageAccountType": "Premium_LRS"
                    },
                    "diskSizeGB": 30
                },
                "dataDisks": []
            },
            "osProfile": data['properties']['osProfile'],
            "networkProfile": data['properties']['networkProfile'],
            "provisioningState": "Creating"
        }
        if 'priority' in origdata['properties']:
            data['properties']['priority'] = origdata['properties']['priority']

        data['zones'] = ["1"]
        self.items.append(data)

        # Add a disk
        disk_data = data.copy()
        disk_data['name'] = 'bionic-azure-' + str(uuid.uuid4())
        disk_data['type'] = "Microsoft.Compute/disks"
        disk_data['id'] = '/'.join(url.path.split('/')[:6] +
                                   [disk_data['type'], disk_data['name']])
        disk_data['properties'] = {"provisioningState": "Succeeded"}
        self.cloud.crud["Microsoft.Compute/disks"].items.append(disk_data)

        resource_group = self.cloud._extract_resource_group(url.path)

        # Add a PIP
        data_nic = (data['properties']['networkProfile']
                    ['networkInterfaceConfigurations'][0])
        data_ip = data_nic['properties']['ipConfigurations'][0]
        pip_url = None
        if 'publicIPAddressConfiguration' in data_ip['properties']:
            data_pip = data_ip['properties']['publicIPAddressConfiguration']
            pip_data = {
                'location': data['location'],
                'tags': data_pip['tags'],
                'sku': {
                    'name': data_pip['sku'],
                },
                'properties': {
                    'publicIPAddressVersion': (data_ip['properties']
                                               ['privateIPAddressVersion']),
                    'publicIPAllocationMethod': (data_pip['properties']
                                                 ['publicIPAllocationMethod']),
                },
            }
            pip_url = (f'/subscriptions/{self.cloud.subscription_id}'
                       f'/resourceGroups/{resource_group}/providers'
                       f'/Microsoft.Network/publicIPAddresses/{name}-v4-abcd')
            self.cloud.crud["Microsoft.Network/publicIPAddresses"]._put(
                pip_data, pip_url)

        # Add a NIC
        nic_data = {
            'location': data['location'],
            'tags': data_nic['tags'],
            'properties': {
                'ipConfigurations': [{
                    'name': data_ip['name'],
                    'properties': {
                        'privateIPAddressVersion': (
                            data_ip['properties']
                            ['privateIPAddressVersion']),
                        'subnet': data_ip['properties']['subnet'],
                    }
                }],
            }
        }
        data['properties']['networkProfile']['networkInterfaces'] =\
            [nic_data.copy()]
        if pip_url:
            (nic_data['properties']['ipConfigurations'][0]['properties']
             ['publicIPAddress']) = {
                'id': pip_url,
                'type': 'Microsoft.Network/publicIPAddresses',
                'properties': {
                    'publicIPAddressVersion': (data_ip['properties']
                                               ['privateIPAddressVersion']),
                },
            }
        nic_url = (f'/subscriptions/{self.cloud.subscription_id}'
                   f'/resourceGroups/{resource_group}/providers'
                   f'/Microsoft.Network/networkInterfaces/{name}-nic-abcd')
        data['properties']['networkProfile']['networkInterfaces'][0]['id'] =\
            nic_url
        self.cloud.crud["Microsoft.Network/networkInterfaces"]._put(
            nic_data, nic_url)

        ret = json.dumps(data)
        # Finish provisioning after return
        data['properties']['provisioningState'] = "Succeeded"
        return (200, {}, ret)

    def delete(self, request):
        url = urllib.parse.urlparse(request.path_url)
        name = url.path.split('/')[-1]
        for item in self.items:
            if item['name'] == name:
                self.items.remove(item)
                return (200, {}, '')
        return (404, {}, json.dumps({
            'error': {
                'message': 'Not Found',
                'code': 'NotFound',
            }}))


class DisksCRUD(CRUDManager):
    name = "Microsoft.Compute/disks"

    def put(self, request):
        data = json.loads(request.body)
        url = urllib.parse.urlparse(request.path_url)
        name = url.path.split('/')[-1]
        data['id'] = url.path
        data['name'] = name
        data['type'] = self.name
        data['properties'] = {
            "provisioningState": "Succeeded",
        }
        self.items.append(data)
        async_url = 'https://management.azure.com/async' + request.path_url
        headers = {'Azure-AsyncOperation': async_url,
                   'Retry-After': '0'}
        return (200, headers, json.dumps(data))

    def post(self, request):
        data = json.loads(request.body)
        url = urllib.parse.urlparse(request.path_url)
        name = url.path.split('/')[-2]
        action = url.path.split('/')[-1]
        if action == 'beginGetAccess':
            async_url = 'https://management.azure.com/async' + request.path_url
            async_url = async_url.replace('/beginGetAccess', '')
            for item in self.items:
                if item['name'] == name:
                    item['accessSAS'] = (
                        'https://management.azure.com/sas/' + name)
        if action == 'endGetAccess':
            async_url = 'https://management.azure.com/async' + request.path_url
            async_url = async_url.replace('/endGetAccess', '')
        headers = {'Azure-AsyncOperation': async_url,
                   'Retry-After': '0'}
        return (200, headers, json.dumps(data))

    def delete(self, request):
        url = urllib.parse.urlparse(request.path_url)
        name = url.path.split('/')[-1]
        async_url = 'https://management.azure.com/async' + request.path_url
        for item in self.items:
            if item['name'] == name:
                self.items.remove(item)
                headers = {'Azure-AsyncOperation': async_url,
                           'Retry-After': '0'}
                return (200, headers, '')
        return (404, {}, json.dumps({
            'error': {
                'message': 'Not Found',
                'code': 'NotFound',
            }}))


class ImagesCRUD(CRUDManager):
    name = "Microsoft.Compute/images"

    def put(self, request):
        data = json.loads(request.body)
        url = urllib.parse.urlparse(request.path_url)
        name = url.path.split('/')[-1]
        data['id'] = url.path
        data['name'] = name
        data['type'] = self.name
        data['properties'] = {
            "provisioningState": "Succeeded",
        }
        self.items.append(data)
        async_url = 'https://management.azure.com/async' + request.path_url
        headers = {'Azure-AsyncOperation': async_url,
                   'Retry-After': '0'}
        return (200, headers, json.dumps(data))

    def delete(self, request):
        url = urllib.parse.urlparse(request.path_url)
        name = url.path.split('/')[-1]
        async_url = 'https://management.azure.com/async' + request.path_url
        for item in self.items:
            if item['name'] == name:
                self.items.remove(item)
                headers = {'Azure-AsyncOperation': async_url,
                           'Retry-After': '0'}
                return (200, headers, '')
        return (404, {}, json.dumps({
            'error': {
                'message': 'Not Found',
                'code': 'NotFound',
            }}))


class FakeAzureFixture(fixtures.Fixture):
    tenant_id = 'b6e5bbf0-ebc0-48a7-9112-ad8829d7edc1'
    subscription_id = '582c3efd-3e57-4073-82c5-d122636fb45c'

    def __init__(self, responses_fixture, quotas):
        super().__init__()
        self.responses = responses_fixture.requests_mock
        self._quotas = quotas

    def _setUp(self):
        self.crud = {}

        self.responses.add(
            responses.POST,
            f'https://login.microsoftonline.com/{self.tenant_id}/oauth2/token',
            json={
                'access_token': 'secret_token',
                'expires_on': time.time() + 600,
            })

        self.responses.add_callback(
            responses.GET,
            ('https://management.azure.com/subscriptions/'
             f'{self.subscription_id}/providers/Microsoft.Compute/skus/'
             '?api-version=2019-04-01'),
            callback=self._get_compute_skus,
            content_type='application/json')

        self.responses.add_callback(
            responses.GET,
            ('https://management.azure.com/subscriptions/'
             f'{self.subscription_id}/providers/Microsoft.Compute/locations/'
             'centralus/usages?api-version=2020-12-01'),
            callback=self._get_compute_usages,
            content_type='application/json')

        async_re = re.compile('https://management.azure.com/async/(.*)')
        self.responses.add_callback(
            responses.GET, async_re,
            callback=self._get_async,
            content_type='application/json')

        sas_re = re.compile('https://management.azure.com/sas/(.*)')
        self.responses.add_callback(
            responses.PUT, sas_re,
            callback=self._put_sas,
            content_type='application/json')

        self._setup_crud(ResourceGroupsCRUD, '2020-06-01',
                         resource_grouped=False)

        self._setup_crud(VirtualMachinesCRUD, '2023-03-01')
        self._setup_crud(NetworkInterfacesCRUD, '2020-11-01')
        self._setup_crud(PublicIPAddressesCRUD, '2020-11-01')
        self._setup_crud(DisksCRUD, '2020-06-30')
        self._setup_crud(ImagesCRUD, '2020-12-01')

    def _setup_crud(self, manager, api_version, resource_grouped=True):
        self.crud[manager.name] = manager(self)

        if resource_grouped:
            rg = 'resourceGroups/(.*?)/providers/'
        else:
            rg = ''

        list_re = re.compile(
            'https://management.azure.com/subscriptions/'
            + f'{self.subscription_id}/'
            + rg + f'{manager.name}/?\\?api-version={api_version}')
        crud_re = re.compile(
            'https://management.azure.com/subscriptions/'
            + f'{self.subscription_id}/'
            + rg + f'{manager.name}/(.+?)\\?api-version={api_version}')
        self.responses.add_callback(
            responses.GET, list_re, callback=self.crud[manager.name].list,
            content_type='application/json')
        self.responses.add_callback(
            responses.GET, crud_re, callback=self.crud[manager.name].get,
            content_type='application/json')
        self.responses.add_callback(
            responses.PUT, crud_re, callback=self.crud[manager.name].put,
            content_type='application/json')
        if hasattr(self.crud[manager.name], 'post'):
            self.responses.add_callback(
                responses.POST, crud_re, callback=self.crud[manager.name].post,
                content_type='application/json')
        if hasattr(self.crud[manager.name], 'delete'):
            self.responses.add_callback(
                responses.DELETE, crud_re,
                callback=self.crud[manager.name].delete,
                content_type='application/json')
        if resource_grouped:
            # Also add list all
            list_re = re.compile(
                'https://management.azure.com/subscriptions/'
                + f'{self.subscription_id}/'
                + f'providers/{manager.name}/?\\?api-version={api_version}')
            self.responses.add_callback(
                responses.GET, list_re, callback=self.crud[manager.name].list,
                content_type='application/json')

    def _extract_resource_group(self, path):
        url = re.compile('/subscriptions/(.*?)/resourceGroups/(.*?)/')
        m = url.match(path)
        return m.group(2)

    def _get_compute_skus(self, request):
        data = {
            'value': [
                {'capabilities': [
                    {'name': 'MaxResourceVolumeMB', 'value': '4096'},
                    {'name': 'OSVhdSizeMB', 'value': '1047552'},
                    {'name': 'vCPUs', 'value': '1'},
                    {'name': 'HyperVGenerations', 'value': 'V1,V2'},
                    {'name': 'MemoryGB', 'value': '0.5'},
                    {'name': 'MaxDataDiskCount', 'value': '2'},
                    {'name': 'LowPriorityCapable', 'value': 'False'},
                    {'name': 'PremiumIO', 'value': 'True'},
                    {'name': 'VMDeploymentTypes', 'value': 'IaaS'},
                    {'name': 'CombinedTempDiskAndCachedIOPS', 'value': '200'},
                    {'name': 'CombinedTempDiskAndCachedReadBytesPerSecond',
                     'value': '10485760'},
                    {'name': 'CombinedTempDiskAndCachedWriteBytesPerSecond',
                     'value': '10485760'},
                    {'name': 'UncachedDiskIOPS', 'value': '160'},
                    {'name': 'UncachedDiskBytesPerSecond',
                     'value': '10485760'},
                    {'name': 'EphemeralOSDiskSupported', 'value': 'True'},
                    {'name': 'EncryptionAtHostSupported', 'value': 'True'},
                    {'name': 'AcceleratedNetworkingEnabled', 'value': 'False'},
                    {'name': 'RdmaEnabled', 'value': 'False'},
                    {'name': 'MaxNetworkInterfaces', 'value': '2'}],
                 'family': 'standardBSFamily',
                 'locationInfo': [
                     {'location': 'centralus',
                      'zoneDetails': [
                          {'Name': ['3', '2', '1'],
                           'capabilities': [
                               {'name': 'UltraSSDAvailable',
                                'value': 'True'}]}],
                      'zones': ['3', '1', '2']}],
                 'locations': ['centralus'],
                 'name': 'Standard_B1ls',
                 'resourceType': 'virtualMachines',
                 'restrictions': [],
                 'size': 'B1ls',
                 'tier': 'Standard'}
            ]
        }
        return (200, {}, json.dumps(data))

    def _get_compute_usages(self, request):
        mgr = self.crud["Microsoft.Compute/virtualMachines"]
        data = {
            'value': [
                {
                    "limit": self._quotas['lowPriorityCores'],
                    "unit": "Count",
                    "currentValue": len(mgr.items),
                    "name": {
                        "value": "lowPriorityCores",
                        "localizedValue": "Total Regional Low-priority vCPUs",
                    }
                }, {
                    "limit": self._quotas['cores'],
                    "unit": "Count",
                    "currentValue": len(mgr.items),
                    "name": {
                        "value": "cores",
                        "localizedValue": "Total Regional vCPUs"
                    }
                }, {
                    "limit": self._quotas['virtualMachines'],
                    "unit": "Count",
                    "currentValue": len(mgr.items),
                    "name": {
                        "value": "virtualMachines",
                        "localizedValue": "Virtual Machines"
                    }
                }
            ]}
        return (200, {}, json.dumps(data))

    def _get_async(self, request):
        path = request.path_url[len('/async'):]
        ret = requests.get('https://management.azure.com' + path)
        data = {
            'status': 'Succeeded',
            'properties': {
                'output': ret.json(),
            }
        }
        return (200, {}, json.dumps(data))

    def _put_sas(self, request):
        return (201, {}, '')



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul.driver import Driver, ConnectionInterface, TriggerInterface
from zuul.driver import SourceInterface, ReporterInterface
from zuul.driver.pagure import pagureconnection
from zuul.driver.pagure import paguremodel
from zuul.driver.pagure import paguresource
from zuul.driver.pagure import pagurereporter
from zuul.driver.pagure import paguretrigger


class PagureDriver(Driver, ConnectionInterface, TriggerInterface,
                   SourceInterface, ReporterInterface):
    name = 'pagure'

    def getConnection(self, name, config):
        return pagureconnection.PagureConnection(self, name, config)

    def getTrigger(self, connection, config=None):
        return paguretrigger.PagureTrigger(self, connection, config)

    def getTriggerEventClass(self):
        return paguremodel.PagureTriggerEvent

    def getSource(self, connection):
        return paguresource.PagureSource(self, connection)

    def getReporter(self, connection, pipeline, config=None,
                    parse_context=None):
        return pagurereporter.PagureReporter(
            self, connection, pipeline, config, parse_context)

    def getTriggerSchema(self):
        return paguretrigger.getSchema()

    def getReporterSchema(self):
        return pagurereporter.getSchema()

    def getRequireSchema(self):
        return paguresource.getRequireSchema()

    def getRejectSchema(self):
        return paguresource.getRejectSchema()



# Copyright 2020 OpenStack Foundation
# Copyright 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import logging


"""Simple Capabilities registry, to be used by Zuul Web."""


class CapabilitiesRegistry(object):

    log = logging.getLogger("Zuul.CapabilitiesRegistry")

    def __init__(self):
        self.capabilities = {}
        self.set_default_capabilities()

    def set_default_capabilities(self):
        self.capabilities['job_history'] = False
        self.capabilities['auth'] = {
            'realms': {},
            'default_realm': None,
        }

    def register_capabilities(self, capability_name, capabilities):
        is_set = self.capabilities.setdefault(capability_name, None)
        if is_set is None:
            action = 'registered'
        else:
            action = 'updated'
        if isinstance(is_set, dict) and isinstance(capabilities, dict):
            self.capabilities[capability_name].update(capabilities)
        else:
            self.capabilities[capability_name] = capabilities
        self.log.debug('Capabilities "%s" %s' % (capability_name, action))


capabilities_registry = CapabilitiesRegistry()



#!/usr/bin/python3


def main():
    raise Exception('This module is broken')


try:
    from ansible.module_utils.basic import *  # noqa
except ImportError:
    pass


if __name__ == '__main__':
    main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os


def get_default(config, section, option, default=None, expand_user=False):
    if config.has_option(section, option):
        # Need to be ensured that we get suitable
        # type from config file by default value
        if isinstance(default, bool):
            value = config.getboolean(section, option)
        elif isinstance(default, int):
            value = config.getint(section, option)
        else:
            value = config.get(section, option)
    else:
        value = default
    if expand_user and value:
        return os.path.expanduser(value)
    return value


def any_to_bool(val):
    val = str(val)
    if val.lower() in ('1', 'on', 'yes', 'true'):
        return True
    return False



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from zuul.lib.result_data import get_artifacts_from_result_data
from zuul.reporter import BaseReporter


class ZuulReporter(BaseReporter):
    def __init__(self, driver, connection, pipeline, config):
        super().__init__(driver, connection, config)
        self.log = logging.getLogger('zuul.reporter')
        self.pipeline = pipeline
        self.image_built = config.get('image-built', False)
        self.image_validated = config.get('image-validated', False)

    def report(self, item, phase1=True, phase2=True):
        if not phase2:
            return
        if self.image_built:
            self.reportImageBuilt(item)
        elif self.image_validated:
            self.reportImageValidated(item)

    def reportImageBuilt(self, item):
        sched = self.driver.sched

        for build in item.current_build_set.getBuilds():
            if not build.job.image_build_name:
                continue
            image_name = build.job.image_build_name
            # Allow the use of an image in a dynamic layout
            layout = item.manager.getLayout(item)
            image = layout.images[image_name]
            ibas = []
            for artifact in get_artifacts_from_result_data(
                    build.result_data,
                    logger=self.log):
                if metadata := artifact.get('metadata'):
                    if metadata.get('type') == 'zuul_image':
                        iba = sched.createImageBuildArtifact(
                            item.manager.tenant.name, image, build,
                            metadata, artifact['url'],
                            self.image_validated)
                        ibas.append(iba)
            for iba in ibas:
                sched.createImageUploads(iba)

    def reportImageValidated(self, item):
        sched = self.driver.sched

        for build in item.current_build_set.getBuilds():
            if not build.job.image_build_name:
                continue
            sched.validateImageUpload(
                item.event.image_upload_uuid,
            )


def getSchema():
    zuul_reporter = {
        'image-built': bool,
        'image-validated': bool,
    }

    return zuul_reporter



# Copyright 2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from concurrent import futures

import fixtures
import grpc
from opentelemetry import trace
from opentelemetry.proto.collector.trace.v1.trace_service_pb2_grpc import (
    TraceServiceServicer,
    add_TraceServiceServicer_to_server
)
from opentelemetry.proto.collector.trace.v1.trace_service_pb2 import (
    ExportTraceServiceResponse,
)


class TraceServer(TraceServiceServicer):
    def __init__(self, fixture):
        super().__init__()
        self.fixture = fixture

    def Export(self, request, context):
        self.fixture.requests.append(request)
        return ExportTraceServiceResponse()


class OTLPFixture(fixtures.Fixture):
    def __init__(self):
        super().__init__()
        self.requests = []
        self.executor = futures.ThreadPoolExecutor(
            thread_name_prefix='OTLPFixture',
            max_workers=10)
        self.server = grpc.server(self.executor)
        add_TraceServiceServicer_to_server(TraceServer(self), self.server)
        self.port = self.server.add_insecure_port('[::]:0')
        # Reset global tracer provider
        trace._TRACER_PROVIDER_SET_ONCE = trace.Once()
        trace._TRACER_PROVIDER = None

    def _setUp(self):
        self.server.start()
        self.addCleanup(self._cleanup)

    def _cleanup(self):
        self.server.stop(None)
        self.server.wait_for_termination()
        self.executor.shutdown()
        # Reset global tracer provider
        trace._TRACER_PROVIDER_SET_ONCE = trace.Once()
        trace._TRACER_PROVIDER = None



# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import voluptuous as v
from zuul.trigger import BaseTrigger
from zuul.driver.gerrit.gerritmodel import GerritEventFilter
from zuul.driver.gerrit import gerritsource
from zuul.driver.util import (
    scalar_or_list,
    to_list,
    make_regex,
    ZUUL_REGEX,
)
from zuul.exceptions import DeprecationWarning


class GerritRequireApprovalDeprecation(DeprecationWarning):
    zuul_error_name = 'Gerrit require-approval Deprecation'
    zuul_error_message = """The 'require-approval' trigger attribute
is deprecated.  Use 'require' instead."""


class GerritRejectApprovalDeprecation(DeprecationWarning):
    zuul_error_name = 'Gerrit reject-approval Deprecation'
    zuul_error_message = """The 'reject-approval' trigger attribute
is deprecated.  Use 'reject' instead."""


class GerritTrigger(BaseTrigger):
    name = 'gerrit'
    log = logging.getLogger("zuul.GerritTrigger")

    def getEventFilters(self, connection_name, trigger_conf,
                        parse_context):
        pcontext = parse_context
        efilters = []
        for trigger in to_list(trigger_conf):
            approvals = {}
            for approval_dict in to_list(trigger.get('approval')):
                for key, val in approval_dict.items():
                    approvals[key] = val
            approval_changes = {}
            for approval_dict in to_list(trigger.get('approval-change')):
                for key, val in approval_dict.items():
                    approval_changes[key] = val
            # Backwards compat for *_filter versions of these args
            attrname = 'comment' if 'comment' in trigger else 'comment_filter'
            with pcontext.confAttr(trigger, attrname) as attr:
                comments = [make_regex(x, pcontext)
                            for x in to_list(attr)]
            attrname = 'email' if 'email' in trigger else 'email_filter'
            with pcontext.confAttr(trigger, attrname) as attr:
                emails = [make_regex(x, pcontext)
                          for x in to_list(attr)]
            attrname = ('username' if 'username' in trigger
                        else 'username_filter')
            with pcontext.confAttr(trigger, attrname) as attr:
                usernames = [make_regex(x, pcontext)
                             for x in to_list(attr)]

            with pcontext.confAttr(trigger, 'event') as attr:
                types = [make_regex(x, pcontext) for x in to_list(attr)]
            with pcontext.confAttr(trigger, 'branch') as attr:
                branches = [make_regex(x, pcontext) for x in to_list(attr)]
            with pcontext.confAttr(trigger, 'ref') as attr:
                refs = [make_regex(x, pcontext) for x in to_list(attr)]
            with pcontext.confAttr(trigger, 'added') as attr:
                added = [make_regex(x, pcontext) for x in to_list(attr)]
            with pcontext.confAttr(trigger, 'removed') as attr:
                removed = [make_regex(x, pcontext) for x in to_list(attr)]

            ignore_deletes = trigger.get('ignore-deletes', True)

            if 'require-approval' in trigger:
                with pcontext.confAttr(trigger, 'require-approval'):
                    pcontext.accumulator.addError(
                        GerritRequireApprovalDeprecation())
            if 'reject-approval' in trigger:
                with pcontext.confAttr(trigger, 'reject-approval'):
                    pcontext.accumulator.addError(
                        GerritRejectApprovalDeprecation())

            f = GerritEventFilter(
                connection_name=connection_name,
                trigger=self,
                types=types,
                branches=branches,
                refs=refs,
                event_approvals=approvals,
                event_approval_changes=approval_changes,
                comments=comments,
                emails=emails,
                usernames=usernames,
                required_approvals=(
                    to_list(trigger.get('require-approval'))
                ),
                reject_approvals=to_list(
                    trigger.get('reject-approval')
                ),
                added=added,
                removed=removed,
                uuid=trigger.get('uuid'),
                scheme=trigger.get('scheme'),
                ignore_deletes=ignore_deletes,
                require=trigger.get('require'),
                reject=trigger.get('reject'),
                debug=trigger.get('debug'),
                parse_context=parse_context,
            )
            efilters.append(f)

        return efilters


def getSchema():
    variable_dict = v.Schema(dict)

    approval = v.Schema({'username': str,
                         'email': str,
                         'older-than': str,
                         'newer-than': str,
                         }, extra=v.ALLOW_EXTRA)

    gerrit_trigger = {
        v.Required('event'):
            scalar_or_list(v.Any('patchset-created',
                                 'draft-published',
                                 'change-abandoned',
                                 'change-restored',
                                 'change-merged',
                                 'comment-added',
                                 'ref-updated',
                                 'pending-check',
                                 'vote-deleted',
                                 'hashtags-changed',
                                 'wip-state-changed')),
        'uuid': str,
        'scheme': str,
        'comment_filter': scalar_or_list(str),
        'comment': scalar_or_list(v.Any(ZUUL_REGEX, str)),
        'email_filter': scalar_or_list(str),
        'email': scalar_or_list(v.Any(ZUUL_REGEX, str)),
        'username_filter': scalar_or_list(str),
        'username': scalar_or_list(v.Any(ZUUL_REGEX, str)),
        'branch': scalar_or_list(v.Any(ZUUL_REGEX, str)),
        'ref': scalar_or_list(v.Any(ZUUL_REGEX, str)),
        'ignore-deletes': bool,
        'approval': scalar_or_list(variable_dict),
        'approval-change': scalar_or_list(variable_dict),
        'require-approval': scalar_or_list(approval),
        'reject-approval': scalar_or_list(approval),
        'added': scalar_or_list(v.Any(ZUUL_REGEX, str)),
        'removed': scalar_or_list(v.Any(ZUUL_REGEX, str)),
        'require': gerritsource.getRequireSchema(),
        'reject': gerritsource.getRejectSchema(),
        'debug': bool,
    }

    return gerrit_trigger



# Copyright 2014 Rackspace Australia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc


class BaseTrigger(object, metaclass=abc.ABCMeta):
    """Base class for triggers.

    Defines the exact public methods that must be supplied."""

    def __init__(self, driver, connection, config=None):
        self.driver = driver
        self.connection = connection
        self.config = config or {}

    @abc.abstractmethod
    def getEventFilters(self, connection_name, trigger_conf,
                        parse_context):
        """Return a list of EventFilter's for the scheduler to match against.
        """

    def postConfig(self, pipeline):
        """Called after config is loaded."""

    def onChangeMerged(self, change, source):
        """Called when a change has been merged."""

    def onChangeEnqueued(self, change, manager, event):
        """Called when a change has been enqueued."""



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import dateutil
import json
import os
import textwrap

from tests.base import AnsibleZuulTestCase


class FunctionalZuulJSONMixIn:
    tenant_config_file = 'config/remote-zuul-json/main.yaml'
    # This should be overriden in child classes.
    ansible_version = 'X'

    def _setUp(self):
        self.fake_nodepool.remote_ansible = True

        ansible_remote = os.environ.get('ZUUL_REMOTE_IPV4')
        self.assertIsNotNone(ansible_remote)

    def _run_job(self, job_name):
        # Keep the jobdir around so we can inspect contents if an
        # assert fails. It will be cleaned up anyway as it is contained
        # in a tmp dir which gets cleaned up after the test.
        self.executor_server.keep_jobdir = True

        # Output extra ansible info so we might see errors.
        self.executor_server.verbose = True
        conf = textwrap.dedent(
            """
            - job:
                name: {job_name}
                run: playbooks/{job_name}.yaml
                ansible-version: {version}
                roles:
                  - zuul: org/common-config
                nodeset:
                  nodes:
                    - name: controller
                      label: whatever

            - project:
                check:
                  jobs:
                    - {job_name}
            """.format(job_name=job_name, version=self.ansible_version))

        file_dict = {'zuul.yaml': conf}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        job = self.getJobFromHistory(job_name)
        return job

    def _get_json_as_text(self, build):
        path = os.path.join(self.jobdir_root, build.uuid,
                            'work', 'logs', 'job-output.json')
        with open(path) as f:
            return f.read()

    def test_no_log(self):
        job = self._run_job('no-log')
        with self.jobLog(job):
            build = self.history[-1]
            self.assertEqual(build.result, 'SUCCESS')

            text = self._get_json_as_text(build)
            self.assertIn('rosebud', text)
            self.assertNotIn('setec', text)

    def test_json_task_action(self):
        job = self._run_job('no-log')
        with self.jobLog(job):
            build = self.history[-1]
            self.assertEqual(build.result, 'SUCCESS')

            text = self._get_json_as_text(build)
            json_result = json.loads(text)
            tasks = json_result[0]['plays'][0]['tasks']
            expected_actions = [
                'debug', 'debug', 'debug', 'copy', 'find',
                'stat', 'debug'
            ]
            for i, expected in enumerate(expected_actions):
                host_result = tasks[i]['hosts']['controller']
                self.assertEqual(expected, host_result['action'])

    def test_json_role_log(self):
        job = self._run_job('json-role')
        with self.jobLog(job):
            build = self.history[-1]
            self.assertEqual(build.result, 'SUCCESS')

            text = self._get_json_as_text(build)
            self.assertIn('json-role', text)

            json_result = json.loads(text)
            role_name = json_result[0]['plays'][0]['tasks'][0]['role']['name']
            self.assertEqual('json-role', role_name)

            role_path = json_result[0]['plays'][0]['tasks'][0]['role']['path']
            self.assertEqual('json-role', os.path.basename(role_path))

    def test_json_time_log(self):
        job = self._run_job('no-log')
        with self.jobLog(job):
            build = self.history[-1]
            self.assertEqual(build.result, 'SUCCESS')

            text = self._get_json_as_text(build)
            # Assert that 'start' and 'end' are part of the result at all
            self.assertIn('start', text)
            self.assertIn('end', text)

            json_result = json.loads(text)
            # Assert that the start and end timestamps are present at the
            # right place in the dictionary
            task = json_result[0]['plays'][0]['tasks'][0]['task']
            task_start_time = task['duration']['start']
            task_end_time = task['duration']['end']

            play = json_result[0]['plays'][0]['play']
            play_start_time = play['duration']['start']
            play_end_time = play['duration']['end']

            # Assert that the start and end timestamps are valid dates
            dateutil.parser.parse(task_start_time)
            dateutil.parser.parse(task_end_time)
            dateutil.parser.parse(play_start_time)
            dateutil.parser.parse(play_end_time)


class TestZuulJSON9(AnsibleZuulTestCase, FunctionalZuulJSONMixIn):
    ansible_version = '9'

    def setUp(self):
        super().setUp()
        self._setUp()


class TestZuulJSON11(AnsibleZuulTestCase, FunctionalZuulJSONMixIn):
    ansible_version = '11'

    def setUp(self):
        super().setUp()
        self._setUp()



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''
This file contains code common to finger log streaming functionality.
The log streamer process within each executor, the finger gateway service,
and the web interface will all make use of this module.
'''

import logging
import os
import pwd
import random
import select
import socket
import socketserver
import ssl
import threading
import time

from zuul.exceptions import StreamingError
from zuul.zk.components import COMPONENT_REGISTRY


log = logging.getLogger("zuul.lib.streamer_utils")


class BaseFingerRequestHandler(socketserver.BaseRequestHandler):
    '''
    Base class for common methods for handling finger requests.
    '''

    MAX_REQUEST_LEN = 1024
    REQUEST_TIMEOUT = 10

    def getCommand(self):
        poll = select.poll()
        bitmask = (select.POLLIN | select.POLLERR |
                   select.POLLHUP | select.POLLNVAL)
        poll.register(self.request, bitmask)
        buffer = b''
        ret = None
        start = time.time()
        while True:
            elapsed = time.time() - start
            timeout = max(self.REQUEST_TIMEOUT - elapsed, 0)
            if not timeout:
                raise Exception("Timeout while waiting for input")
            for fd, event in poll.poll(timeout):
                if event & select.POLLIN:
                    x = self.request.recv(self.MAX_REQUEST_LEN)
                    if not x:
                        # This will cause the caller to quietly shut down
                        raise BrokenPipeError
                    buffer += x
                else:
                    raise Exception("Received error event")
            if len(buffer) >= self.MAX_REQUEST_LEN:
                raise Exception("Request too long")
            try:
                ret = buffer.decode('utf-8')
                x = ret.find('\n')
                if x > 0:
                    # rstrip to remove any other unnecessary chars (e.g. \r)
                    return ret[:x].rstrip()
            except UnicodeDecodeError:
                pass


class CustomThreadingTCPServer(socketserver.ThreadingTCPServer):
    '''
    Custom version that allows us to drop privileges after port binding.
    '''

    def __init__(self, *args, **kwargs):
        # NOTE(pabelanger): Set up address_family for socketserver based on the
        # fingergw.listen_address setting in zuul.conf.
        # param tuple args[0]: The address and port to bind to for
        # socketserver.
        server_address = args[0]
        address_family = None
        for res in socket.getaddrinfo(
                server_address[0], server_address[1], 0, self.socket_type):
            if res[0] == socket.AF_INET6:
                # If we get an IPv6 address, break our loop and use that first.
                address_family = res[0]
                break
            elif res[0] == socket.AF_INET:
                address_family = res[0]

        # Check to see if getaddrinfo failed.
        if not address_family:
            raise Exception("getaddrinfo returns an empty list")

        self.address_family = address_family
        self.user = kwargs.pop('user', None)
        self.pid_file = kwargs.pop('pid_file', None)
        self.server_ssl_key = kwargs.pop('server_ssl_key', None)
        self.server_ssl_cert = kwargs.pop('server_ssl_cert', None)
        self.server_ssl_ca = kwargs.pop('server_ssl_ca', None)
        socketserver.ThreadingTCPServer.__init__(self, *args, **kwargs)

    def change_privs(self):
        '''
        Drop our privileges to another user.
        '''
        if os.getuid() != 0:
            return

        pw = pwd.getpwnam(self.user)

        # Change owner on our pid file so it can be removed by us after
        # dropping privileges. May not exist if not a daemon.
        if self.pid_file and os.path.exists(self.pid_file):
            os.chown(self.pid_file, pw.pw_uid, pw.pw_gid)

        os.setgroups([])
        os.setgid(pw.pw_gid)
        os.setuid(pw.pw_uid)
        os.umask(0o022)

    def server_bind(self):
        '''
        Overridden from the base class to allow address reuse and to drop
        privileges after binding to the listening socket.
        '''
        self.allow_reuse_address = True
        socketserver.ThreadingTCPServer.server_bind(self)
        if self.user:
            self.change_privs()

    def server_close(self):
        '''
        Overridden from base class to shutdown the socket immediately.
        '''
        try:
            self.socket.shutdown(socket.SHUT_RD)
            self.socket.close()
        except socket.error as e:
            # If it's already closed, don't error.
            if e.errno == socket.EBADF:
                return
            raise

    def process_request(self, request, client_address):
        '''
        Overridden from the base class to name the thread.
        '''
        t = threading.Thread(target=self.process_request_thread,
                             name='socketserver_Thread',
                             args=(request, client_address))
        t.daemon = self.daemon_threads
        t.start()

    def get_request(self):
        sock, addr = super().get_request()

        if all([self.server_ssl_key, self.server_ssl_cert,
                self.server_ssl_ca]):
            context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            context.load_cert_chain(self.server_ssl_cert, self.server_ssl_key)
            context.load_verify_locations(self.server_ssl_ca)
            context.verify_mode = ssl.CERT_REQUIRED
            sock = context.wrap_socket(sock, server_side=True)
        return sock, addr


def getJobLogStreamAddress(executor_api, uuid, source_zone,
                           tenant_name=None):
    """
    Looks up the log stream address for the given build UUID.

    Try to find the build request for the given UUID in ZooKeeper
    by searching through all available zones. If a build request
    was found we use the worker information to build the log stream
    address.
    """
    build_request = executor_api.getRequest(uuid)

    if build_request is None:
        raise StreamingError("Build not found")

    if tenant_name is not None and build_request.tenant_name != tenant_name:
        # Intentionally the same error as above to avoid leaking
        # out-of-tenant build information.
        raise StreamingError("Build not found")

    worker_info = build_request.worker_info
    if not worker_info:
        raise StreamingError("Build did not start yet")

    worker_zone = worker_info.get("zone")
    job_log_stream_address = {}
    if worker_zone and source_zone != worker_zone:
        info = _getFingerGatewayInZone(worker_zone)
        if info:
            job_log_stream_address['server'] = info.hostname
            job_log_stream_address['port'] = info.public_port
            job_log_stream_address['use_ssl'] = info.use_ssl
            log.debug('Source (%s) and worker (%s) zone '
                      'are different, routing via %s:%s',
                      source_zone, worker_zone,
                      info.hostname, info.public_port)
        else:
            log.warning('Source (%s) and worker (%s) zone are different'
                        'but no fingergw in target zone found. '
                        'Falling back to direct connection.',
                        source_zone, worker_zone)
    else:
        log.debug('Source (%s) or worker zone (%s) undefined '
                  'or equal, no routing is needed.',
                  source_zone, worker_zone)

    if 'server' not in job_log_stream_address:
        job_log_stream_address['server'] = worker_info["hostname"]
        job_log_stream_address['port'] = worker_info["log_port"]

    return job_log_stream_address


def _getFingerGatewayInZone(zone):
    registry = COMPONENT_REGISTRY.registry
    gws = [gw for gw in registry.all('fingergw') if gw.zone == zone]
    if gws:
        return random.choice(gws)
    return None



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import os
import time
import yaml

from tests.base import ZuulTestCase, simple_layout


class TestGitDriver(ZuulTestCase):
    config_file = 'zuul-git-driver.conf'
    tenant_config_file = 'config/git-driver/main.yaml'

    def setUp(self):
        super(TestGitDriver, self).setUp()
        self.git_connection = self.scheds.first.sched.connections\
            .getSource('git').connection

    def setup_config(self, config_file: str):
        config = super(TestGitDriver, self).setup_config(config_file)
        config.set('connection git', 'baseurl', self.upstream_root)
        return config

    def test_basic(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        # Check that we have the git source for common-config and the
        # gerrit source for the project.
        self.assertEqual('git', list(tenant.config_projects)[0].source.name)
        self.assertEqual('common-config', list(tenant.config_projects)[0].name)
        self.assertEqual('gerrit',
                         list(tenant.untrusted_projects)[0].source.name)
        self.assertEqual('org/project',
                         list(tenant.untrusted_projects)[0].name)

        # The configuration for this test is accessed via the git
        # driver (in common-config), rather than the gerrit driver, so
        # if the job runs, it worked.
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(A.reported, 1)

    def test_config_refreshed(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual(A.reported, 1)
        self.assertEqual(self.history[0].name, 'project-test1')

        # Update zuul.yaml to force a tenant reconfiguration
        path = os.path.join(self.upstream_root, 'common-config', 'zuul.yaml')
        with open(path, 'r') as f:
            config = yaml.safe_load(f)
        change = {
            'name': 'org/project',
            'check': {
                'jobs': [
                    'project-test2'
                ]
            }
        }
        config[4]['project'] = change
        files = {'zuul.yaml': yaml.dump(config)}
        self.addCommitToRepo(
            'common-config', 'Change zuul.yaml configuration', files)

        # Wait for the tenant reconfiguration to happen
        count = self.waitForEvent()
        self.waitUntilSettled()

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 2)
        self.assertEqual(A.reported, 1)
        # We make sure the new job has run
        self.assertEqual(self.history[1].name, 'project-test2')

        # Let's stop the git Watcher to let us merge some changes commits
        # We want to verify that config changes are detected for commits
        # on the range oldrev..newrev
        self.scheds.first.sched.connections.getSource('git').connection\
            .watcher_thread._pause = True
        # Add a config change
        change = {
            'name': 'org/project',
            'check': {
                'jobs': [
                    'project-test1'
                ]
            }
        }
        config[4]['project'] = change
        files = {'zuul.yaml': yaml.dump(config)}
        self.addCommitToRepo(
            'common-config', 'Change zuul.yaml configuration', files)
        # Add two other changes
        self.addCommitToRepo(
            'common-config', 'Adding f1',
            {'f1': "Content"})
        self.addCommitToRepo(
            'common-config', 'Adding f2',
            {'f2': "Content"})
        # Restart the git watcher
        self.scheds.first.sched.connections.getSource('git').connection\
            .watcher_thread._pause = False

        # Wait for the tenant reconfiguration to happen
        self.waitForEvent(count)
        self.waitUntilSettled()

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 3)
        self.assertEqual(A.reported, 1)
        # We make sure the new job has run
        self.assertEqual(self.history[2].name, 'project-test1')

    def ensure_watcher_has_context(self):
        # Make sure watcher have read initial refs shas
        delay = 0.1
        max_delay = 1
        while not self.git_connection.watcher_thread.projects_refs:
            time.sleep(delay)
            max_delay -= delay
            if max_delay <= 0:
                raise Exception("Timeout waiting for initial read")
        return self.git_connection.watcher_thread._event_count

    def waitForEvent(self, initial_count=0):
        delay = 0.1
        max_delay = 5
        while self.git_connection.watcher_thread._event_count <= initial_count:
            time.sleep(delay)
            max_delay -= delay
            if max_delay <= 0:
                raise Exception("Timeout waiting for event")
        return self.git_connection.watcher_thread._event_count

    @simple_layout('layouts/basic-git.yaml', driver='git')
    def test_ref_updated_event(self):
        count = self.ensure_watcher_has_context()
        # Add a commit to trigger a ref-updated event
        self.addCommitToRepo(
            'org/project', 'A change for ref-updated', {'f1': 'Content'})
        # Wait for the git watcher to detect the ref-update event
        self.waitForEvent(count)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('post-job').result)

    @simple_layout('layouts/basic-git.yaml', driver='git')
    def test_ref_created(self):
        count = self.ensure_watcher_has_context()
        # Tag HEAD to trigger a ref-updated event
        self.addTagToRepo(
            'org/project', 'atag', 'HEAD')
        # Wait for the git watcher to detect the ref-update event
        self.waitForEvent(count)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 1)
        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('tag-job').result)

    @simple_layout('layouts/basic-git.yaml', driver='git')
    def test_ref_deleted(self):
        count = self.ensure_watcher_has_context()
        # Delete default tag init to trigger a ref-updated event
        self.delTagFromRepo(
            'org/project', 'init')
        # Wait for the git watcher to detect the ref-update event
        self.waitForEvent(count)
        self.waitUntilSettled()
        # Make sure no job as run as ignore-delete is True by default
        self.assertEqual(len(self.history), 0)

    @simple_layout('layouts/basic-git.yaml', driver='git')
    def test_get_project_branch_sha(self):
        # Exercise this method since it's only called from timer
        # triggers
        source = self.scheds.first.sched.connections.getSource('git')
        project = source.getProject('org/project')
        self.assertIsNotNone(source.getProjectBranchSha(project, 'master'))



# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
import urllib
import logging
import voluptuous as vs
from urllib.parse import urlparse
from zuul.source import BaseSource
from zuul.model import Project
from zuul.driver.gerrit.gerritmodel import GerritRefFilter
from zuul.driver.util import (
    scalar_or_list,
    ZUUL_REGEX,
)
from zuul.lib.dependson import find_dependency_headers
from zuul.zk.change_cache import ChangeKey


def noneOrStr(val):
    if val is None:
        return None
    return str(val)


class GerritSource(BaseSource):
    name = 'gerrit'
    log = logging.getLogger("zuul.source.Gerrit")

    def __init__(self, driver, connection, config=None):
        hostname = connection.canonical_hostname
        super(GerritSource, self).__init__(driver, connection,
                                           hostname, config)
        prefix_ui = urlparse(self.connection.baseurl).path
        if prefix_ui:
            prefix_ui = prefix_ui.lstrip('/').rstrip('/')
            prefix_ui += '/'

        self.change_re = re.compile(
            r"/%s(\#\/c\/|c\/.*\/\+\/)?(\d+)[\w]*" % prefix_ui)

    def getRefSha(self, project, ref):
        return self.connection.getRefSha(project, ref)

    def isMerged(self, change, head=None):
        return self.connection.isMerged(change, head)

    def canMerge(self, change, allow_needs, event=None, allow_refresh=False):
        # Gerrit changes have no volatile data that cannot be updated via
        # events and thus needs not to act on allow_refresh.
        return self.connection.canMerge(change, allow_needs, event=event)

    def postConfig(self):
        pass

    def getChangeKey(self, event):
        connection_name = self.connection.connection_name
        if event.change_number:
            return ChangeKey(connection_name, None,
                             'GerritChange',
                             str(event.change_number),
                             noneOrStr(event.patch_number))
        revision = f'{event.oldrev}..{event.newrev}'
        if event.ref and event.ref.startswith('refs/tags/'):
            tag = event.ref[len('refs/tags/'):]
            return ChangeKey(connection_name, event.project_name,
                             'Tag', tag, revision)
        if event.ref and not event.ref.startswith('refs/'):
            # Pre 2.13 Gerrit ref-updated events don't have branch prefixes.
            return ChangeKey(connection_name, event.project_name,
                             'Branch', event.ref, revision)
        if event.ref and event.ref.startswith('refs/heads/'):
            # From the timer trigger or Post 2.13 Gerrit
            branch = event.ref[len('refs/heads/'):]
            return ChangeKey(connection_name, event.project_name,
                             'Branch', branch, revision)
        if event.ref:
            # catch-all ref (ie, not a branch or head)
            return ChangeKey(connection_name, event.project_name,
                             'Ref', event.ref, revision)
        self.log.warning("Unable to format change key for %s" % (event,))

    def getChange(self, change_key, refresh=False, event=None):
        return self.connection.getChange(change_key, refresh=refresh,
                                         event=event)

    def getProjectBranchSha(self, project, branch_name):
        return (
            self.connection.getRefSha(project, f'refs/heads/{branch_name}')
            or None
        )

    def getChangeByURL(self, url, event):
        try:
            parsed = urllib.parse.urlparse(url)
        except ValueError:
            return None
        path = parsed.path
        if parsed.fragment:
            path += '#' + parsed.fragment
        m = self.change_re.match(path)
        if not m:
            return None
        try:
            change_no = int(m.group(2))
        except ValueError:
            return None
        query = "change:%s" % (change_no,)
        results = self.connection.simpleQuery(query, event=event)
        if not results:
            return None
        change_key = ChangeKey(self.connection.connection_name, None,
                               'GerritChange',
                               str(results[0].number),
                               str(results[0].current_patchset))
        change = self.connection._getChange(change_key, event=event)
        return change

    def getChangesDependingOn(self, change, projects, tenant):
        changes = []
        if not change.uris:
            return changes
        queries = set()
        for uri in change.uris:
            queries.add('message:{Depends-On: %s}' % uri)
        query = '(' + ' OR '.join(queries) + ')'
        results = self.connection.simpleQuery(query)
        self.log.debug('%s possible depending changes found', len(results))
        seen = set()
        for result in results:
            for match in find_dependency_headers(result.message):
                found = False
                for uri in change.uris:
                    if uri in match:
                        found = True
                        break
                if not found:
                    continue
                key = (result.number, result.current_patchset)
                if key in seen:
                    continue
                seen.add(key)
                change_key = ChangeKey(self.connection.connection_name, None,
                                       'GerritChange',
                                       str(result.number),
                                       str(result.current_patchset))
                change = self.connection._getChange(change_key)
                changes.append(change)
        return changes

    def useDependenciesByTopic(self):
        return bool(self.connection.submit_whole_topic)

    def getChangesByTopic(self, topic, event=None, changes=None, history=None):
        if not topic:
            return []

        if changes is None:
            changes = {}

        if history is None:
            history = []
        history.append(topic)

        query = 'status:open topic:"%s"' % topic
        results = self.connection.simpleQuery(query)
        for result in results:
            change_key = ChangeKey(self.connection.connection_name, None,
                                   'GerritChange',
                                   str(result.number),
                                   str(result.current_patchset))
            if change_key in changes:
                continue

            change = self.connection._getChange(change_key, event=event)
            changes[change_key] = change

        # Convert to list here because the recursive call can mutate
        # the set.
        for change in list(changes.values()):
            for git_change_ref in change.git_needs_changes:
                change_key = ChangeKey.fromReference(git_change_ref)
                if change_key in changes:
                    continue
                git_change = self.getChange(change_key, event=event)
                if not git_change.topic or git_change.topic in history:
                    continue
                self.getChangesByTopic(
                    git_change.topic, event, changes, history)
        return list(changes.values())

    def getCachedChanges(self):
        yield from self.connection._change_cache

    def getProject(self, name):
        p = self.connection.getProject(name)
        if not p:
            p = Project(name, self)
            self.connection.addProject(p)
        return p

    def getProjectOpenChanges(self, project):
        return self.connection.getProjectOpenChanges(project)

    def getProjectDefaultMergeMode(self, project, valid_modes=None):
        # The gerrit jgit merge operation is most closely approximated
        # by "git merge -s resolve", so we return that as the default
        # for the Gerrit driver.
        return 'merge-resolve'

    def getProjectDefaultBranch(self, project, tenant, min_ltime=-1):
        # We have to return something here, so try to get it from the
        # cache, and if that fails, return the Zuul default.
        try:
            default_branch = self.connection.getProjectDefaultBranch(
                project, tenant, min_ltime)
        except Exception:
            default_branch = None
        if default_branch is None:
            return super().getProjectDefaultBranch(project, tenant, min_ltime)
        return default_branch

    def getProjectBranches(self, project, tenant, min_ltime=-1):
        return self.connection.getProjectBranches(project, tenant, min_ltime)

    def getProjectBranchCacheLtime(self):
        return self.connection._branch_cache.ltime

    def getGitUrl(self, project):
        return self.connection.getGitUrl(project)

    def _getGitwebUrl(self, project, sha=None):
        return self.connection._getGitwebUrl(project, sha)

    def getRequireFilters(self, config, parse_context):
        f = GerritRefFilter.requiresFromConfig(
            self.connection.connection_name,
            config, parse_context)
        return [f]

    def getRejectFilters(self, config, parse_context):
        f = GerritRefFilter.rejectFromConfig(
            self.connection.connection_name,
            config, parse_context)
        return [f]

    def getRefForChange(self, change):
        partial = str(change).zfill(2)[-2:]
        return "refs/changes/%s/%s/.*" % (partial, change)

    def setChangeAttributes(self, change, **attrs):
        return self.connection.updateChangeAttributes(change, **attrs)


approval = vs.Schema({'username': str,
                      'email': str,
                      'older-than': str,
                      'newer-than': str,
                      }, extra=vs.ALLOW_EXTRA)


def getRequireSchema():
    require = {'approval': scalar_or_list(approval),
               'open': bool,
               'current-patchset': bool,
               'wip': bool,
               'hashtags': scalar_or_list(vs.Any(ZUUL_REGEX, str)),
               'status': scalar_or_list(str)}
    return require


def getRejectSchema():
    reject = {'approval': scalar_or_list(approval),
              'open': bool,
              'current-patchset': bool,
              'wip': bool,
              'hashtags': scalar_or_list(vs.Any(ZUUL_REGEX, str)),
              'status': scalar_or_list(str)}
    return reject



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import logging
import signal
import testtools

import zuul.cmd


class TestStackDump(testtools.TestCase):
    def setUp(self):
        super(TestStackDump, self).setUp()
        self.log_fixture = self.useFixture(
            fixtures.FakeLogger(level=logging.DEBUG))

    def test_stack_dump_logs(self):
        "Test that stack dumps end up in logs."

        zuul.cmd.stack_dump_handler(signal.SIGUSR2, None)
        self.assertIn("Thread", self.log_fixture.output)
        self.assertIn("MainThread", self.log_fixture.output)
        self.assertIn("test_stack_dump_logs", self.log_fixture.output)
        # Call it again in case yappi or objgraph are installed in the
        # venv so that we turn them back off.
        zuul.cmd.stack_dump_handler(signal.SIGUSR2, None)



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul import model
from zuul.lib.logutil import get_annotated_logger
from zuul.manager.shared import SharedQueuePipelineManager


class DependentPipelineManager(SharedQueuePipelineManager):
    """PipelineManager for handling interrelated Changes.

    The DependentPipelineManager puts Changes that share a Pipeline
    into a shared :py:class:`~zuul.model.ChangeQueue`. It then processes them
    using the Optimistic Branch Prediction logic with Nearest Non-Failing Item
    reparenting algorithm for handling errors.
    """
    changes_merge = True
    type = 'dependent'

    def __init__(self, *args, **kwargs):
        super(DependentPipelineManager, self).__init__(*args, **kwargs)

    def constructChangeQueue(self, queue_name):
        p = self.pipeline
        return model.ChangeQueue.new(
            self.current_context,
            manager=self,
            window=p.window,
            window_floor=p.window_floor,
            window_ceiling=p.window_ceiling,
            window_increase_type=p.window_increase_type,
            window_increase_factor=p.window_increase_factor,
            window_decrease_type=p.window_decrease_type,
            window_decrease_factor=p.window_decrease_factor,
            name=queue_name)

    def getNodePriority(self, item, change):
        return item.queue.queue.index(item)

    def areChangesReadyToBeEnqueued(self, changes, event,
                                    warnings=None, debug=False):
        log = get_annotated_logger(self.log, event)
        for change in changes:
            source = change.project.source
            can_merge = source.canMerge(change, self.getSubmitAllowNeeds(),
                                        event=event)
            if not can_merge:
                msg = (
                    f"Change {change._id()} "
                    f"in project {change.project} "
                    "can not be merged"
                )
                if isinstance(can_merge, model.FalseWithReason):
                    msg += f" due to: {can_merge}"
                log.debug("  " + msg)
                if debug and warnings is not None:
                    warnings.append(msg)
                return False
        return True

    def getNonMergeableCycleChanges(self, item):
        """Return changes in the cycle that do not fulfill
        the pipeline's ready criteria."""
        changes = []
        for change in item.changes:
            source = change.project.source
            if not source.canMerge(
                change,
                self.getSubmitAllowNeeds(),
                event=item.event,
                allow_refresh=True,
            ):
                log = get_annotated_logger(self.log, item.event)
                log.debug("Change %s can no longer be merged", change)
                changes.append(change)
        return changes

    def enqueueChangesBehind(self, change, event, quiet, ignore_requirements,
                             change_queue, history=None,
                             dependency_graph=None):
        log = get_annotated_logger(self.log, event)
        history = history if history is not None else []

        log.debug("Checking for changes needing %s:" % change)
        if not isinstance(change, model.Change):
            log.debug("  %s does not support dependencies" % type(change))
            return

        # for project in change_queue, project.source get changes, then dedup.
        projects = [self.tenant.getProject(pcn)[1] for pcn, _ in
                    change_queue.project_branches]
        sources = {p.source for p in projects}

        needed_by_changes = self.resolveChangeReferences(
            change.getNeededByChanges())
        log.debug("  Previously known following changes: %s",
                  needed_by_changes)
        seen = set(needed_by_changes)
        for source in sources:
            log.debug("  Checking source: %s", source)
            relevant_projects = [p for p in projects if p.source == source]
            for c in source.getChangesDependingOn(change,
                                                  relevant_projects,
                                                  self.tenant):
                if c not in seen:
                    seen.add(c)
                    needed_by_changes.append(c)

        log.debug("  Updated following changes: %s", needed_by_changes)

        to_enqueue = []
        change_dependencies = dependency_graph.get(change, [])
        for other_change in needed_by_changes:
            if other_change in change_dependencies:
                # Only consider the change if it is not part of a cycle, as
                # cycle changes will otherwise be partially enqueued without
                # any error handling
                self.log.debug(
                    "    Skipping change %s due to dependency cycle",
                    other_change
                )
                continue

            with self.getChangeQueue(other_change,
                                     event) as other_change_queue:
                if other_change_queue != change_queue:
                    log.debug("  Change %s in project %s can not be "
                              "enqueued in the target queue %s" %
                              (other_change, other_change.project,
                               change_queue))
                    continue
            source = other_change.project.source
            if source.canMerge(other_change, self.getSubmitAllowNeeds(),
                               event=event):
                log.debug("  Change %s needs %s and is ready to merge",
                          other_change, change)
                to_enqueue.append(other_change)

        if not to_enqueue:
            log.debug("  No changes need %s" % change)

        for other_change in to_enqueue:
            self.addChange(other_change, event, quiet=quiet,
                           ignore_requirements=ignore_requirements,
                           change_queue=change_queue, history=history,
                           dependency_graph=dependency_graph)

    def enqueueChangesAhead(self, changes, event, quiet, ignore_requirements,
                            change_queue, history=None, dependency_graph=None,
                            warnings=None, debug=False):
        log = get_annotated_logger(self.log, event)

        history = history if history is not None else []
        for change in changes:
            if hasattr(change, 'number'):
                history.append(change)
            else:
                # Don't enqueue dependencies ahead of a non-change ref.
                return True

        abort, needed_changes = self.getMissingNeededChanges(
            changes, change_queue, event,
            dependency_graph=dependency_graph,
            warnings=warnings, debug=debug)
        if abort:
            return False

        if not needed_changes:
            return True
        log.debug("  Changes %s must be merged ahead of %s",
                  needed_changes, change)
        for needed_change in needed_changes:
            # If the change is already in the history, but the change also has
            # a git level dependency, we need to enqueue it before the current
            # change.
            if (needed_change not in history or
                    needed_change.cache_key in change.git_needs_changes):
                r = self.addChange(needed_change, event, quiet=quiet,
                                   ignore_requirements=ignore_requirements,
                                   change_queue=change_queue, history=history,
                                   dependency_graph=dependency_graph,
                                   warnings=warnings, debug=debug)
                if not r:
                    return False
        return True

    def getMissingNeededChanges(self, changes, change_queue, event,
                                dependency_graph=None, warnings=None,
                                item=None, debug=False):
        log = get_annotated_logger(self.log, event)
        changes_needed = []
        abort = False

        # Return true if okay to proceed enqueing this change,
        # false if the change should not be enqueued.
        for change in changes:
            log.debug("Checking for changes needed by %s:" % change)
            if not isinstance(change, model.Change):
                log.debug("  %s does not support dependencies", type(change))
                continue
            needed_changes = dependency_graph.get(change)
            if not needed_changes:
                log.debug("  No changes needed")
                continue
            # Ignore supplied change_queue
            with self.getChangeQueue(change, event) as change_queue:
                for needed_change in needed_changes:
                    log.debug("  Change %s needs change %s:" % (
                        change, needed_change))
                    if needed_change.is_merged:
                        log.debug("  Needed change is merged")
                        continue
                    with self.getChangeQueue(needed_change,
                                             event) as needed_change_queue:
                        if needed_change_queue != change_queue:
                            msg = ("Change %s in project %s does not "
                                   "share a change queue with %s "
                                   "in project %s" %
                                   (needed_change._id(),
                                    needed_change.project,
                                    change.number,
                                    change.project))
                            log.debug("  " + msg)
                            if warnings is not None:
                                warnings.append(msg)
                            changes_needed.append(needed_change)
                            abort = True
                    if not needed_change.is_current_patchset:
                        msg = (
                            f"Needed change {needed_change._id()} "
                            f"in project {needed_change.project} is not "
                            "the current patchset"
                        )
                        log.debug("  " + msg)
                        changes_needed.append(needed_change)
                        if debug and warnings is not None:
                            warnings.append(msg)
                        abort = True
                    if needed_change in changes:
                        log.debug("  Needed change is in cycle")
                        continue
                    if self.isChangeAlreadyInQueue(
                            needed_change, change_queue, event, item):
                        log.debug("  Needed change is already "
                                  "ahead in the queue")
                        continue
                    can_merge = needed_change.project.source.canMerge(
                        needed_change, self.getSubmitAllowNeeds(),
                        event=event)
                    if can_merge:
                        msg = (
                            f"Change {needed_change._id()} "
                            f"in project {needed_change.project} is needed "
                        )
                        log.debug("  " + msg)
                        if needed_change not in changes_needed:
                            changes_needed.append(needed_change)
                        continue
                    else:
                        # The needed change can't be merged.
                        msg = (
                            f"Change {needed_change._id()} "
                            f"in project {needed_change.project} is needed "
                            "but can not be merged"
                        )
                        if isinstance(can_merge, model.FalseWithReason):
                            msg += f" due to: {can_merge}"
                        log.debug("  " + msg)
                        if debug and warnings is not None:
                            warnings.append(msg)
                        changes_needed.append(needed_change)
                        abort = True
        return abort, changes_needed

    def getFailingDependentItems(self, item):
        failing_items = set()
        for change in item.changes:
            if not isinstance(change, model.Change):
                continue
            needs_changes = change.getNeedsChanges(
                self.useDependenciesByTopic(change.project))
            if not needs_changes:
                continue
            for needed_change in self.resolveChangeReferences(needs_changes):
                needed_item = self.getItemForChange(needed_change)
                if not needed_item:
                    continue
                if needed_item is item:
                    continue
                if needed_item.current_build_set.failing_reasons:
                    failing_items.add(needed_item)
        return failing_items

    def dequeueItem(self, item, quiet=False):
        super(DependentPipelineManager, self).dequeueItem(item, quiet)
        # If this was a dynamic queue from a speculative change,
        # remove the queue (if empty)
        if item.queue.dynamic:
            if not item.queue.queue:
                self.state.removeQueue(item.queue)
